# Copyright (c) SI-Analytics. All rights reserved.
from .codebase import *  # noqa F403
from .core import *  # noqa F403
from .tune import *  # noqa F403
from .utils import *  # noqa F403
from .version import __version__, version_info

__all__ = ['__version__', 'version_info']



# Copyright (c) SI-Analytics. All rights reserved.
from typing import Dict, List, Optional, Union

from ray.tune.result import DEFAULT_METRIC
from ray.tune.search.nevergrad import NevergradSearch as _NevergradSearch

from .builder import SEARCHERS

try:
    import nevergrad as ng
    from nevergrad.optimization import Optimizer
    from nevergrad.optimization.base import ConfiguredOptimizer
    Parameter = ng.p.Parameter
    from nevergrad.optimization.optimizerlib import \
        registry as optimizer_registry
except ImportError:
    ng = None
    Optimizer = None
    ConfiguredOptimizer = None
    Parameter = None
    optimizer_registry = dict()


@SEARCHERS.register_module(force=True)
class NevergradSearch(_NevergradSearch):
    """Search with Nevergrad."""

    def __init__(self,
                 optimizer: str = 'OnePlusOne',
                 space: Optional[Union[Dict, Parameter]] = None,
                 metric: Optional[str] = None,
                 mode: Optional[str] = None,
                 points_to_evaluate: Optional[List[Dict]] = None,
                 num_workers: int = 1,
                 budget: Optional[int] = None,
                 **kwargs) -> None:
        """Initialize NevergradSearch.

        Args:
            optimizer (str): The optimizer. Defaults to 'OnePlusOne'.
            space (Optional[Union[Dict, Parameter]]):
                The space to search. Defaults to None.
            metric (Optional[str]):
                Performance evaluation metrics.. Defaults to None.
            mode (Optional[str]):
                Determines whether objective is
                minimizing or maximizing the metric attribute.
                Defaults to None.
            points_to_evaluate (Optional[List[Dict]]):
                Initial parameter suggestions to be run first.
                Defaults to None.
            num_workers (int):
                The number of evaluations
                which will be run in parallel at once.
                Defaults to 1.
            budget (Optional[int]):
                The number of allowed evaluations.
                Defaults to None.
        """
        assert optimizer in optimizer_registry, (
            f'{optimizer} is not registered')
        optimizer = optimizer_registry[optimizer]
        self._budget = budget
        self._num_workers = num_workers
        super(NevergradSearch, self).__init__(
            optimizer=optimizer,
            space=space,
            metric=metric,
            mode=mode,
            points_to_evaluate=points_to_evaluate,
            **kwargs)

    def _setup_nevergrad(self) -> None:
        """setup Nevergrad optimizer."""

        if self._opt_factory:
            self._nevergrad_opt = self._opt_factory(
                parametrization=self._space,
                budget=self._budget,
                num_workers=self._num_workers)

        # nevergrad.tell internally minimizes, so "max" => -1
        if self._mode == 'max':
            self._metric_op = -1.
        elif self._mode == 'min':
            self._metric_op = 1.

        if self._metric is None and self._mode:
            # If only a mode was passed, use anonymous metric
            self._metric = DEFAULT_METRIC

        if hasattr(self._nevergrad_opt, 'instrumentation'):  # added in v0.2.0
            if self._nevergrad_opt.instrumentation.kwargs:
                if self._nevergrad_opt.instrumentation.args:
                    raise ValueError(
                        'Instrumented optimizers should use kwargs only')
                if self._parameters is not None:
                    raise ValueError('Instrumented optimizers should provide '
                                     'None as parameter_names')
            else:
                if self._parameters is None:
                    raise ValueError('Non-instrumented optimizers should have '
                                     'a list of parameter_names')
                if len(self._nevergrad_opt.instrumentation.args) != 1:
                    raise ValueError(
                        'Instrumented optimizers should use kwargs only')
        if self._parameters is not None and (self._nevergrad_opt.dimension !=
                                             len(self._parameters)):
            raise ValueError('len(parameters_names) must match optimizer '
                             'dimension for non-instrumented optimizers')

        if self._points_to_evaluate:
            # Nevergrad is LIFO, so we add the points to evaluate in reverse
            # order.
            for i in range(len(self._points_to_evaluate) - 1, -1, -1):
                self._nevergrad_opt.suggest(self._points_to_evaluate[i])

    def add_evaluated_point(
        self,
        parameters: Dict,
        value: float,
        error: bool = False,
        pruned: bool = False,
        intermediate_values: Optional[List[float]] = None,
        trial_id: str = None,
    ) -> None:
        """Add evaluated point to Nevergrad optimizer.

        Args:
            parameters (Dict): The parameters.
            value (float): The value.
            error (bool):
                Whether the point was evaluated in error. Defaults to False.
            pruned (bool):
                Whether the point was pruned. Defaults to False.
            intermediate_values (Optional[List[float]]):
                The intermediate values. Defaults to None.
            trial_id (str): The trial id. Defaults to None.
        """

        candidate = self._nevergrad_opt.parametrization.spawn_child(
            new_value=parameters)

        self._nevergrad_opt.tell(candidate, self._metric_op * value)



# Copyright (c) SI-Analytics. All rights reserved.
from typing import List

from ray.tune.experiment import Trial
from ray.tune.integration.mlflow import \
    MLflowLoggerCallback as _MLflowLoggerCallback
from ray.tune.integration.mlflow import logger
from ray.tune.utils.util import is_nan_or_inf

from .builder import CALLBACKS


@CALLBACKS.register_module()
class MLflowLoggerCallback(_MLflowLoggerCallback):
    """Custom MLflow Logger to automatically log Tune results and config to
    MLflow. The main differences from the original MLflow Logger are:

        1. Bind multiple runs into a parent run in the form of nested run.
        2. Log artifacts of the best trial to the parent run.

    Refer to https://github.com/ray-project/ray/blob/ray-1.9.1/python/ray/tune/integration/mlflow.py for details.  # noqa E501

    Args:
        metric (str): Key for trial info to order on.
        mode (str): One of [min, max]. Defaults to ``self.default_mode``.
        scope (str): One of [all, last, avg, last-5-avg, last-10-avg].
            If `scope=last`, only look at each trial's final step for
            `metric`, and compare across trials based on `mode=[min,max]`.
            If `scope=avg`, consider the simple average over all steps
            for `metric` and compare across trials based on
            `mode=[min,max]`. If `scope=last-5-avg` or `scope=last-10-avg`,
            consider the simple average over the last 5 or 10 steps for
            `metric` and compare across trials based on `mode=[min,max]`.
            If `scope=all`, find each trial's min/max score for `metric`
            based on `mode`, and compare trials based on `mode=[min,max]`.
        filter_nan_and_inf (bool): If True, NaN or infinite values
            are disregarded and these trials are never selected as
            the best trial. Default: True.
        **kwargs: kwargs for original ``MLflowLoggerCallback``
    """

    def __init__(self,
                 metric: str = None,
                 mode: str = None,
                 scope: str = 'last',
                 filter_nan_and_inf: bool = True,
                 **kwargs):
        super(MLflowLoggerCallback, self).__init__(**kwargs)
        self.metric = metric
        if mode and mode not in ['min', 'max']:
            raise ValueError('`mode` has to be None or one of [min, max]')
        self.mode = mode
        if scope not in ['all', 'last', 'avg', 'last-5-avg', 'last-10-avg']:
            raise ValueError(
                'ExperimentAnalysis: attempting to get best trial for '
                "metric {} for scope {} not in [\"all\", \"last\", \"avg\", "
                "\"last-5-avg\", \"last-10-avg\"]. "
                "If you didn't pass a `metric` parameter to `tune.run()`, "
                'you have to pass one when fetching the best trial.'.format(
                    self.metric, scope))
        self.scope = scope if scope != 'all' else mode
        self.filter_nan_and_inf = filter_nan_and_inf

    def setup(self, *args, **kwargs):
        """In addition to create `mlflow` experiment, create a parent run to
        bind multiple trial runs."""
        super().setup(*args, **kwargs)

        self.client = self.mlflow_util._get_client()
        self.experiment_id = self.mlflow_util.experiment_id
        self.parent_run = self.client.create_run(
            experiment_id=self.experiment_id, tags=self.tags)

    def log_trial_start(self, trial: 'Trial'):
        """Overrides `log_trial_start` of original `MLflowLoggerCallback` to
        set the parent run ID.

        Args:
            trial (Trial): :class:`ray.tune.experiment.trial.Trial`
        """
        # Create run if not already exists.
        if trial not in self._trial_runs:

            # Set trial name in tags
            tags = self.tags.copy()
            tags['trial_name'] = str(trial)
            tags['mlflow.parentRunId'] = self.parent_run.info.run_id

            run = self.client.create_run(
                experiment_id=self.experiment_id, tags=tags)
            self._trial_runs[trial] = run.info.run_id

        run_id = self._trial_runs[trial]

        # Log the config parameters.
        config = trial.config

        for key, value in config.items():
            self.client.log_param(run_id=run_id, key=key, value=value)

    def on_experiment_end(self, trials: List['Trial'], **info):
        """Overrides `Callback` of `Callback` to copy a best trial to parent
        run. Called after experiment is over and all trials have concluded.

        Args:
            trials (List[Trial]): List of trials.
            **info: Kwargs dict for forward compatibility.
        """
        if not self.metric or not self.mode:
            return

        best_trial, best_score = None, None
        for trial in trials:
            if self.metric not in trial.metric_analysis:
                continue

            score = trial.metric_analysis[self.metric][self.scope]
            if self.filter_nan_and_inf and is_nan_or_inf(score):
                continue

            best_score = best_score or score
            if self.mode == 'max' and score >= best_score or (
                    self.mode == 'min' and score <= best_score):
                best_trial, best_score = trial, score

        if best_trial is None:
            logger.warning(
                'Could not find best trial. Did you pass the correct `metric` '
                'parameter?')
            return

        if best_trial not in self._trial_runs:
            return

        # Copy the run of best trial to parent run.
        run_id = self._trial_runs[best_trial]
        run = self.client.get_run(run_id)
        parent_run_id = self.parent_run.info.run_id

        for key, value in run.data.params.items():
            self.client.log_param(run_id=parent_run_id, key=key, value=value)

        for key, value in run.data.metrics.items():
            self.client.log_metric(run_id=parent_run_id, key=key, value=value)

        if self.should_save_artifact:
            self.client.log_artifacts(
                parent_run_id, local_dir=best_trial.logdir)

        self.client.set_terminated(run_id=parent_run_id, status='FINISHED')



# Copyright (c) SI-Analytics. All rights reserved.
from .context import ContextManager
from .hooks import *  # noqa F403
from .launch import DistributedTorchLauncher
from .rewriters import REWRITERS, build_rewriter

__all__ = [
    'ContextManager',
    'REWRITERS',
    'build_rewriter',
    'DistributedTorchLauncher',
]



# Copyright (c) SI-Analytics. All rights reserved.
from mmengine.registry import Registry
from ray.tune.experiment import Trial

NAME_CREATOR = Registry('name creator')


@NAME_CREATOR.register_module()
def trial_id(trial: Trial) -> str:
    return trial.trial_id


@NAME_CREATOR.register_module()
def experiment_tag(trial: Trial) -> str:
    return trial.experiment_tag



# Copyright (c) SI-Analytics. All rights reserved.
from .callbacks import *  # noqa F403
from .schedulers import *  # noqa F403
from .searchers import *  # noqa F403
from .spaces import *  # noqa F403
from .stoppers import *  # noqa F403
from .tuner import Tuner

__all__ = ['Tuner']



# Copyright (c) SI-Analytics. All rights reserved.
#
# Brought from https://github.com/facebookresearch/nevergrad/blob/main/nevergrad/functions/corefuncs.py # noqa E501
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
import argparse

import numpy as np
import ray
from mmengine.config import Config

from .blackbox import BlackBoxTask
from .builder import TASKS


@TASKS.register_module()
class DiscreteTestFunction(BlackBoxTask):
    """Test functions for discrete optimization.

    Returns a classical discrete function for test, in the domain
    {0,1,...,arity-1}^d
    """

    @staticmethod
    def onemax(x: np.ndarray, arity: int) -> float:
        """onemax(x) is the most classical case of discrete functions, adapted
        to minimization. It is originally designed for lists of bits. It just
        counts the number of 1,

        and returns len(x) - number of ones.
        However, the present function perturbs the location of the optimum,
        so that tests can not be easily biased by a wrong initialization.
        So the optimum,
        instead of being located at (1,1,...,1),
        is located at (0,1,2,...,arity-1,0,1,2,...).

        Args:
            x (np.ndarray): input vector
            arity (int): The number of categories a discrete variable can take

        Returns:
            float: output of the function.
        """
        diff = np.round(x) - (np.arange(len(x)) % arity)
        return float(np.sum(diff != 0))

    @staticmethod
    def leadingones(x: np.ndarray, arity: int) -> float:
        """leadingones is the second most classical discrete function, adapted
        for minimization. Before perturbation of the location of the optimum as
        above,

        it returns len(x) - number of initial 1. I.e.
        leadingones([0 1 1 1]) = 4,
        leadingones([1 1 1 1]) = 0,
        leadingones([1 0 0 0]) = 3.

        Args:
            x (np.ndarray): input vector
            arity (int): The number of categories a discrete variable can take

        Returns:
            float: output of the function.
        """
        diff = np.round(x) - (np.arange(len(x)) % arity)
        nonzeros = np.nonzero(diff)[0]
        return float(len(x) - nonzeros[0] if nonzeros.size else 0)

    @staticmethod
    def jump(x: np.ndarray, arity: int) -> float:
        """A variant of the jump function based on the principle that local
        descent does not work.

        Args:
            x (np.ndarray): input vector
            arity (int): The number of categories a discrete variable can take

        Returns:
            float: output of the function.
        """
        n = len(x)
        m = n // 4
        o = n - DiscreteTestFunction.onemax(x, arity)
        if o == n or o <= n - m:
            return n - m - o
        return o

    def run(self, *, args: argparse.Namespace, **kwargs) -> None:
        """Run the task.

        Args:
            args (argparse.Namespace): The arguments.
        """
        cfg = Config.fromfile(args.config)
        arity = cfg.get('arity', 2)
        func = getattr(self, cfg.get('func', 'onemax'))

        inputs = []
        for k, v in cfg.items():
            if k.startswith('_variable'):
                inputs.append(v)
        inputs = np.array(inputs)
        return ray.tune.report(result=func(inputs, arity))



# Copyright (c) SI-Analytics. All rights reserved.
from typing import Callable

import ray
import torch
from ray import tune
from ray.train._internal.utils import get_address_and_port

from siatune.utils import set_env_vars


class DistributedTorchLauncher:
    """Distributed Training Launcher for PyTorch.

    This class provides a simple
    interface for launching distributed
    training jobs with PyTorch.
    It uses the Ray framework to manage
    the distributed environment
    and to parallelize the training process.

    Args:
        num_cpus_per_worker (int, optional):
            Number of CPU cores to allocate per worker.
            Defaults to 1.
        num_workers (int, optional):
            Number of workers to launch.
            Defaults to 1.

    Attributes:
        num_workers (int):
            Number of workers launched.
        num_cpus_per_worker (int):
            Number of CPU cores allocated per worker.
        num_gpus_per_worker (int):
            Number of GPUs available per worker.
        resources (tune.PlacementGroupFactory):
            Ray placement group for resource allocation.
    """

    def __init__(self, num_cpus_per_worker: int = 1, num_workers: int = 1):
        self.num_workers = num_workers
        self.num_cpus_per_worker = num_cpus_per_worker
        self.num_gpus_per_worker = int(torch.cuda.is_available())
        self.resources = tune.PlacementGroupFactory([
            dict(CPU=self.num_cpus_per_worker, GPU=self.num_gpus_per_worker),
        ] * num_workers)

    def launch(self, func: Callable, *args, **kwargs):
        """Launch the distributed training job.

        Args:
            func (Callable):
                Training function to be executed
                in the distributed environment.
            *args:
                Positional arguments to be passed
                to the training function.
            **kwargs:
                Keyword arguments to be passed
                to the training function.
        """
        addr, port = get_address_and_port()

        def job(rank):
            set_env_vars(rank, self.num_workers, addr, port)
            func(*args, **kwargs)

        remote_job = ray.remote(job).options(
            num_cpus=self.num_cpus_per_worker,
            num_gpus=self.num_gpus_per_worker)
        futures = [
            remote_job.remote(rank) for rank in range(1, self.num_workers)
        ]
        job(rank=0)
        # To facilitate session sharing,
        # the master job will be initiated by the trainer.
        ray.get(futures)



# Copyright (c) SI-Analytics. All rights reserved.
import argparse
from abc import ABCMeta
from typing import Callable, Sequence

from .base import BaseTask
from .builder import TASKS


@TASKS.register_module()
class BlackBoxTask(BaseTask, metaclass=ABCMeta):
    """A general wrapping class for optimizing black box systems (without any
    knowledge of its internal workings.

    Its implementation is "opaque").
    """

    def parse_args(self, args: Sequence[str]) -> argparse.Namespace:
        """Define and parse the necessary arguments for the task.

        Args:
            args (Sequence[str]): The args.
        Returns:
            argparse.Namespace: The parsed args.
        """

        parser = argparse.ArgumentParser(description='black box')
        return parser.parse_args(args)

    def create_trainable(self) -> Callable:
        """Get ray trainable task.

        Returns:
            Callable: The Ray trainable task.
        """

        return self.context_aware_run



# Copyright (c) SI-Analytics. All rights reserved.
import argparse
import tempfile
from os import path as osp
from typing import Dict

from ray.air import session

from siatune.utils import dump_cfg
from .base import BaseRewriter
from .builder import REWRITERS


@REWRITERS.register_module()
class Dump(BaseRewriter):

    arg_name: str = 'config'
    raw_arg_idx: int = 0
    """Dump the configs in the context as a file."""

    def __init__(self, key: str):
        """Inintialize the Dump class.

        Args:
            key (str): The key in the context.
        """
        self.key = key

    def _dump(self, cfg: Dict) -> str:
        tmpdir = tempfile.gettempdir()
        trial_id = session.get_trial_id()
        dmp_path = osp.join(tmpdir, f'{trial_id}.py')
        dump_cfg(cfg, dmp_path)
        return dmp_path

    def __call__(self, context: Dict) -> Dict:
        """Dump the configs in the context.

        Args:
            context (Dict): The context to be rewritten.

        Returns:
            Dict: The context after rewriting.
        """
        is_parsed = isinstance(context['args'], argparse.Namespace)
        dmp_path = self._dump(context.pop(self.key))
        if is_parsed:
            setattr(context['args'], self.arg_name, dmp_path)
        else:
            context['args'][self.raw_arg_idx] = dmp_path
        return context



# Copyright (c) SI-Analytics. All rights reserved.
import argparse
from argparse import REMAINDER
from os import path as osp

import ray
from mmengine.config.config import Config, DictAction

from siatune.apis import log_analysis
from siatune.tune import Tuner
from siatune.utils import register_all_modules


def parse_args():
    parser = argparse.ArgumentParser(description='Tune a model')
    parser.add_argument('config', help='tune config file path')
    parser.add_argument(
        '--work-dir', default=None, help='the dir to save logs and models')
    parser.add_argument(
        '--resume', default=None, help='the experiment path to resume')
    parser.add_argument(
        '--cfg-options',
        nargs='+',
        action=DictAction,
        help='override some settings in the used config, the key-value pair '
        'in xxx=yyy format will be merged into config file. If the value to '
        'be overwritten is a list, it should be like key="[a,b]" or key=a,b '
        'It also allows nested list/tuple values, e.g. key="[(a,b),(c,d)]" '
        'Note that the quotation marks are necessary and that no white space '
        'is allowed.')
    parser.add_argument(
        '--address',
        default=None,
        help='the address of the ray cluster to connect to',
    )
    parser.add_argument(
        '--num-cpus',
        default=None,
        type=int,
        help='number of CPUs the user wishes to assign',
    )
    parser.add_argument(
        '--num-gpus',
        default=None,
        type=int,
        help='number of GPUs the user wishes to assign',
    )
    parser.add_argument(
        '--num-workers', type=int, default=1, help='number of workers to use')
    parser.add_argument(
        '--num-gpus-per-worker',
        type=int,
        default=1,
        help='number of gpus each worker uses.',
    )
    parser.add_argument(
        '--num-cpus-per-worker',
        type=int,
        default=1,
        help='number of gpus each worker uses.',
    )
    parser.add_argument(
        '--trainable-args',
        nargs=REMAINDER,
        type=str,
        help='Rest from the trainable process.',
    )
    args = parser.parse_args()
    return args


def main():
    args = parse_args()

    # register all modules in siatune into the registries
    # do not init the default scope here because it will be init in the runner
    register_all_modules(init_default_scope=False)

    # load config
    cfg = Config.fromfile(args.config)
    if args.cfg_options is not None:
        cfg.merge_from_dict(args.cfg_options)

    # work_dir is determined in this priority: CLI > segment in file > filename
    if args.work_dir is not None:
        # update configs according to CLI args if args.work_dir is not None
        cfg.work_dir = args.work_dir
    elif cfg.get('work_dir', None) is None:
        # use config filename as default work_dir if cfg.work_dir is None
        cfg.work_dir = osp.join('./work_dirs',
                                osp.splitext(osp.basename(args.config))[0])

    # resume is determined in this priority: CLI > segment in file
    if args.resume is not None:
        cfg.resume = args.resume

    # task arguments are determined in this priority: CLI > segment in file
    if args.trainable_args is not None:
        cfg.task.args = args.trainable_args

    # set resource
    cfg.task.num_workers = args.num_workers
    cfg.task.num_cpus_per_worker = args.num_cpus_per_worker
    cfg.task.num_gpus_per_worker = args.num_gpus_per_worker

    # init ray
    if not ray.is_initialized():
        ray.init(
            address=args.address,
            num_cpus=args.num_cpus,
            num_gpus=args.num_gpus)
    assert ray.is_initialized()

    # start tuning
    tuner = Tuner.from_cfg(cfg)
    results = tuner.tune()

    log_analysis(
        results, log_dir=osp.join(tuner.work_dir, tuner.experiment_name))



# Copyright (c) SI-Analytics. All rights reserved.
#
# Brought from https://github.com/facebookresearch/nevergrad/blob/main/nevergrad/functions/corefuncs.py # noqa E501
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
import argparse
from math import exp, sqrt, tanh

import numpy as np
import ray
from mmengine.config import Config

from .blackbox import BlackBoxTask
from .builder import TASKS


def _styblinksitang(x: np.ndarray, noise: float) -> float:
    """Classical function for testing noisy optimization.

    Args:
        x (np.ndarray): input vector.
        noise (float): noise level.
    Returns:
        float: output.
    """
    x2 = x**2
    val = x2.dot(x2) + np.sum(5 * x - 16 * x2)
    # return a positive value for maximization
    return float(39.16599 * len(x) + 0.5 * val +  # noqa W504
                 noise * np.random.normal(size=val.shape))


def _step(s: float, eps: float = 1e-8) -> float:
    """Step function.

    Args:
        s (float): input.
        eps (float): small number to avoid numerical errors.
    Returns:
        float: output.
    """
    if s == 0:
        s += eps
    return np.nan_to_num(np.exp(np.round(np.log(s))))


@TASKS.register_module()
class ContinuousTestFunction(BlackBoxTask):
    """Continuous test function."""

    @staticmethod
    def delayedsphere(x: np.ndarray) -> float:
        """Delayed sphere function.

        Args:
            x (np.ndarray): input vector.
        Returns:
            float: output.
        """
        return float(np.sum(x**2))

    @staticmethod
    def sphere(x: np.ndarray) -> float:
        """The most classical continuous optimization testbed.

        Args:
            x (np.ndarray): input vector.
        Returns:
            float: output.
        """
        assert x.ndim == 1
        return float(x.dot(x))

    # TODO: change translated the sphere function as partial function
    @staticmethod
    def sphere1(x: np.ndarray) -> float:
        """Translated sphere function.

        Args:
            x (np.ndarray): input vector.
        Returns:
            float: output.
        """
        return ContinuousTestFunction.sphere(x - 1.0)

    @staticmethod
    def sphere2(x: np.ndarray) -> float:
        """A bit more translated sphere function.

        Args:
            x (np.ndarray): input vector.
        Returns:
            float: output.
        """
        return ContinuousTestFunction.sphere(x - 2.0)

    @staticmethod
    def sphere4(x: np.ndarray) -> float:
        """Even more translated sphere function.

        Args:
            x (np.ndarray): input vector.
        Returns:
            float: output.
        """
        return ContinuousTestFunction.sphere(x - 4.0)

    @staticmethod
    def maxdeceptive(x: np.ndarray) -> float:
        """Max deceptive function.

        Args:
            x (np.ndarray): input vector.
        Returns:
            float: output.
        """
        dec = 3 * x**2 - (2 / (3**(x - 2)**2 + 0.1))
        return float(np.max(dec))

    @staticmethod
    def sumdeceptive(x: np.ndarray) -> float:
        """Sum deceptive function.

        Args:
            x (np.ndarray): input vector.
        Returns:
            float: output.
        """
        dec = 3 * x**2 - (2 / (3**(x - 2)**2 + 0.1))
        return float(np.sum(dec))

    @staticmethod
    def altcigar(x: np.ndarray) -> float:
        """Similar to cigar, but variables in inverse order.

        E.g. for pointing out algorithms not invariant to the order of
        variables.

        Args:
            x (np.ndarray): input vector.
        Returns:
            float: output.
        """
        return float(
            x[-1])**2 + 1000000.0 * ContinuousTestFunction.sphere(x[:-1])

    @staticmethod
    def discus(x: np.ndarray) -> float:
        """Only one variable is very penalized.

        Args:
            x (np.ndarray): input vector.
        Returns:
            float: output.
        """
        return ContinuousTestFunction.sphere(
            x[1:]) + 1000000.0 * float(x[0])**2

    @staticmethod
    def cigar(x: np.ndarray) -> float:
        """Classical example of ill conditioned function.

        The other classical example is ellipsoid.

        Args:
            x (np.ndarray): input vector.
        Returns:
            float: output.
        """
        return float(
            x[0])**2 + 1000000.0 * ContinuousTestFunction.sphere(x[1:])

    @staticmethod
    def bentcigar(x: np.ndarray) -> float:
        """Classical example of ill conditioned function, but bent.

        Args:
            x (np.ndarray): input vector.
        Returns:
            float: output.
        """
        y = np.asarray([
            x[i]**(
                1 + 0.5 * np.sqrt(x[i]) *  # noqa W504
                (i - 1)(len(x) - 1)) if x[i] > 0.0 else x[i]
            for i in range(len(x))
        ])
        return float(
            y[0])**2 + 1000000.0 * ContinuousTestFunction.sphere(y[1:])

    @staticmethod
    def multipeak(x: np.ndarray) -> float:
        """Inspired by M.

        Gallagher's Gaussian peaks function.

        Args:
            x (np.ndarray): input vector.
        Returns:
            float: output.
        """
        v = 10000.0
        for a in range(101):
            x_ = np.asarray([np.cos(a + np.sqrt(i)) for i in range(len(x))])
            v = min(v,
                    a / 101.0 + np.exp(ContinuousTestFunction.sphere(x - x_)))
        return v

    @staticmethod
    def altellipsoid(x: np.ndarray) -> float:
        """Similar to Ellipsoid, but variables in inverse order.

        E.g. for pointing out algorithms not invariant to the order of
        variables.

        Args:
            x (np.ndarray): input vector.
        Returns:
            float: output.
        """
        return ContinuousTestFunction.ellipsoid(x[::-1])

    @staticmethod
    def stepellipsoid(x: np.ndarray) -> float:
        """Classical example of ill conditioned function.

        But we add a 'step', i.e. we set the gradient to zero everywhere.
        Compared to some existing testbeds, we decided to have infinitely many
        steps.

        Args:
            x (np.ndarray): input vector.
        Returns:
            float: output.
        """
        dim = x.size
        weights = 10**np.linspace(0, 6, dim)
        return float(_step(weights.dot(x**2)))

    @staticmethod
    def ellipsoid(x: np.ndarray) -> float:
        """Classical example of ill conditioned function.

        The other classical example is cigar.

        Args:
            x (np.ndarray): input vector.
        Returns:
            float: output.
        """
        dim = x.size
        weights = 10**np.linspace(0, 6, dim)
        return float(weights.dot(x**2))

    @staticmethod
    def rastrigin(x: np.ndarray) -> float:
        """Classical multimodal function.

        Args:
            x (np.ndarray): input vector.
        Returns:
            float: output.
        """
        cosi = float(np.sum(np.cos(2 * np.pi * x)))
        return float(10 * (len(x) - cosi) + ContinuousTestFunction.sphere(x))

    @staticmethod
    def bucherastrigin(x: np.ndarray) -> float:
        """Classical multimodal function.

        No box-constraint penalization here.

        Args:
            x (np.ndarray): input vector.
        Returns:
            float: output.
        """
        s = np.asarray([
            x[i] * (10 if x[i] > 0.0 and i % 2 else 1) *  # noqa W504
            (10**((i - 1) / (2 * (len(x) - 1)))) for i in range(len(x))
        ])
        cosi = float(np.sum(np.cos(2 * np.pi * s)))
        return float(10 * (len(x) - cosi) + ContinuousTestFunction.sphere(s))

    @staticmethod
    def doublelinearslope(x: np.ndarray) -> float:
        """We decided to use two linear slopes rather than having a constraint
        artificially added for not having the optimum at infinity.

        Args:
            x (np.ndarray): input vector.
        Returns:
            float: output.
        """
        return float(np.abs(np.sum(x)))

    @staticmethod
    def stepdoublelinearslope(x: np.ndarray) -> float:
        """Double Lenear Slope with step.

        Args:
            x (np.ndarray): input vector.
        Returns:
            float: output.
        """
        return _step(np.abs(np.sum(x)))

    @staticmethod
    def hm(x: np.ndarray, eps: float = 1e-8) -> float:
        """New multimodal function (proposed for Nevergrad).

        Args:
            x (np.ndarray): input vector.
            eps (float): small number to avoid numerical errors.
        Returns:
            float: output.
        """
        if not np.all(x):
            x += eps
        return float((x**2).dot(1.1 + np.cos(1.0 / x)))

    @staticmethod
    def rosenbrock(x: np.ndarray) -> float:
        """Rosenbrock function is a non-convex function, introduced by Howard
        H. Rosenbrock in 1960.

        Args:
            x (np.ndarray): input vector.
        Returns:
            float: output.
        """
        x_m_1 = x[:-1] - 1
        x_diff = x[:-1]**2 - x[1:]
        return float(100 * x_diff.dot(x_diff) + x_m_1.dot(x_m_1))

    @staticmethod
    def ackley(x: np.ndarray) -> float:
        """Ackley function is a non-convex function used as a performance test
        problem for optimization algorithms. It was proposed by David Ackley in
        his 1987 PhD dissertation.

        Args:
            x (np.ndarray): input vector.
        Returns:
            float: output.
        """
        dim = x.size
        sum_cos = np.sum(np.cos(2 * np.pi * x))
        return (
            -20.0 * exp(-0.2 * sqrt(ContinuousTestFunction.sphere(x) / dim))
            -  # noqa W504
            exp(sum_cos / dim) + 20 + exp(1))

    @staticmethod
    def schwefel_1_2(x: np.ndarray) -> float:
        """Schwefel's function is deceptive in that the global minimum is
        geometrically distant, over the parameter space, from the next best
        local minima. Therefore, the search algorithms are potentially prone to
        convergence in the wrong direction.

        Args:
            x (np.ndarray): input vector.
        Returns:
            float: output.
        """
        cx = np.cumsum(x)
        return ContinuousTestFunction.sphere(cx)

    @staticmethod
    def griewank(x: np.ndarray) -> float:
        """Multimodal function, often used in Bayesian optimization.

        Args:
            x (np.ndarray): input vector.
        Returns:
            float: output.
        """
        part1 = ContinuousTestFunction.sphere(x)
        part2 = np.prod(np.cos(x / np.sqrt(1 + np.arange(len(x)))))
        return 1 + (float(part1) / 4000.0) - float(part2)

    @staticmethod
    def deceptiveillcond(x: np.ndarray) -> float:
        """An extreme ill conditioned functions.

        Most algorithms fail on this. The condition number increases to
        infinity as we get closer to the optimum.

        Args:
            x (np.ndarray): input vector.
        Returns:
            float: output.
        """
        assert len(x) >= 2
        return float(
            max(
                np.abs(np.arctan(x[1] / x[0])),
                np.sqrt(x[0]**2.0 + x[1]**2.0),
                1.0 if x[0] > 0 else 0.0,
            ) if x[0] != 0.0 else float('inf'))

    @staticmethod
    def deceptivepath(x: np.ndarray) -> float:
        """A function which needs following a long path.

        Most algorithms fail on this. The path becomes thiner as we get closer
        to the optimum.

        Args:
            x (np.ndarray): input vector.
        Returns:
            float: output.
        """
        assert len(x) >= 2
        distance = np.sqrt(x[0]**2 + x[1]**2)
        if distance == 0.0:
            return 0.0
        angle = np.arctan(x[0] / x[1]) if x[1] != 0.0 else np.pi / 2.0
        invdistance = (1.0 / distance) if distance > 0.0 else 0.0
        if np.abs(np.cos(invdistance) - angle) > 0.1:
            return 1.0
        return float(distance)

    @staticmethod
    def deceptivemultimodal(x: np.ndarray) -> float:
        """Infinitely many local optima, as we get closer to the optimum.

        Args:
            x (np.ndarray): input vector.
        Returns:
            float: output.
        """
        assert len(x) >= 2
        distance = np.sqrt(x[0]**2 + x[1]**2)
        if distance == 0.0:
            return 0.0
        angle = np.arctan(x[0] / x[1]) if x[1] != 0.0 else np.pi / 2.0
        invdistance = int(1.0 / distance) if distance > 0.0 else 0.0
        if np.abs(np.cos(invdistance) - angle) > 0.1:
            return 1.0
        return float(distance)

    @staticmethod
    def lunacek(x: np.ndarray) -> float:
        """Known as the bior double-Rastrigin function, is a hybrid function
        consisting of a Rastrigin and a double-sphere part.

        Args:
            x (np.ndarray): input vector.
        Returns:
            float: output.
        """
        problemDimensions = len(x)
        s = 1.0 - (1.0 / (2.0 * np.sqrt(problemDimensions + 20.0) - 8.2))
        mu1 = 2.5
        mu2 = -np.sqrt(abs((mu1**2 - 1.0) / s))
        firstSum = 0.0
        secondSum = 0.0
        thirdSum = 0.0
        for i in range(problemDimensions):
            firstSum += (x[i] - mu1)**2
            secondSum += (x[i] - mu2)**2
            thirdSum += 1.0 - np.cos(2 * np.pi * (x[i] - mu1))
        return min(firstSum,
                   1.0 * problemDimensions + secondSum) + 10 * thirdSum

    @staticmethod
    def genzcornerpeak(x: np.ndarray) -> float:
        """One of the Genz functions, originally used in integration, tested in
        optim because why not.

        Args:
            x (np.ndarray): input vector.
        Returns:
            float: output.
        """
        value = float(1 + np.mean(np.tanh(x)))
        if value == 0:
            return float('inf')
        return value**(-len(x) - 1)

    @staticmethod
    def minusgenzcornerpeak(x: np.ndarray) -> float:
        """One of the Genz functions, originally used in integration, tested in
        optim because why not.

        Args:
            x (np.ndarray): input vector.
        Returns:
            float: output.
        """
        return -ContinuousTestFunction.genzcornerpeak(x)

    @staticmethod
    def genzgaussianpeakintegral(x: np.ndarray) -> float:
        """One of the Genz functions, originally used in integration, tested in
        optim because why not.

        Args:
            x (np.ndarray): input vector.
        Returns:
            float: output.
        """
        return exp(-ContinuousTestFunction.sphere(x) / 4.0)

    @staticmethod
    def minusgenzgaussianpeakintegral(x: np.ndarray) -> float:
        """One of the Genz functions, originally used in integration, tested in
        optim because why not.

        Args:
            x (np.ndarray): input vector.
        Returns:
            float: output.
        """
        return -ContinuousTestFunction.genzgaussianpeakintegral(x)

    @staticmethod
    def slope(x: np.ndarray) -> float:
        """The summation of the vector.

        Args:
            x (np.ndarray): input vector.
        Returns:
            float: output.
        """
        return sum(x)

    @staticmethod
    def linear(x: np.ndarray) -> float:
        """The hyperbolic tangent of first element.

        Args:
            x (np.ndarray): input vector.

        Returns:
            float: output.
        """
        return tanh(x[0])

    @staticmethod
    def st0(x: np.ndarray) -> float:
        """Styblinksitang function with 0 noise.

        Args:
            x (np.ndarray): input vector.
        Returns:
            float: output.
        """
        return _styblinksitang(x, 0)

    @staticmethod
    def st1(x: np.ndarray) -> float:
        """Styblinksitang function with noise 1.

        Args:
            x (np.ndarray): input vector.
        Returns:
            float: output.
        """
        return _styblinksitang(x, 1)

    @staticmethod
    def st10(x: np.ndarray) -> float:
        """Styblinksitang function with noise 10.

        Args:
            x (np.ndarray): input vector.
        Returns:
            float: output.
        """
        return _styblinksitang(x, 10)

    @staticmethod
    def st100(x: np.ndarray) -> float:
        """Styblinksitang function with noise 100.

        Args:
            x (np.ndarray): input vector.
        Returns:
            float: output.
        """
        return _styblinksitang(x, 100)

    def run(self, *, args: argparse.Namespace, **kwargs) -> None:
        """Run the task.

        Args:
            args (argparse.Namespace): The arguments.
        """
        cfg = Config.fromfile(args.config)
        func = getattr(self, cfg.get('func', 'sphere'))
        inputs = []
        for k, v in cfg.items():
            if k.startswith('_variable'):
                inputs.append(v)
        inputs = np.array(inputs)

        ray.tune.report(result=func(inputs))



# Copyright (c) SI-Analytics. All rights reserved.
from collections import defaultdict
from typing import Dict

from ray.tune import Stopper

from .builder import STOPPERS


@STOPPERS.register_module()
class EarlyDroppingStopper(Stopper):
    """Drop Bad Trials."""

    def __init__(self,
                 metric: str,
                 mode: str,
                 metric_threshold: float,
                 grace_period: int = 0) -> None:
        """Initialize early dropping stopper.

        Args:
            metric (str): Metric to optimize.
            mode (str):
                Determines whether objective is
                minimizing or maximizing the metric attribute.
            metric_threshold (float): The threshold for early stopping.
            grace_period (int):
                Only stop trials at least this old in time. Defaults to 0.
        """

        if mode not in ['min', 'max']:
            raise ValueError('mode must be either "min" or "max".')

        self._metric = metric
        self._mode = mode
        self._metric_threshold = metric_threshold
        self._iter = defaultdict(lambda: 0)
        self._grace_period = grace_period

    def __call__(self, trial_id: str, result: Dict) -> bool:
        """Check if trial should be stopped.

        Args:
            trial_id (str): Trial ID.
            result (Dict): Trial result.

        Returns:
            bool: Whether to stop trial.
        """

        self._iter[trial_id] += 1
        metric_result = result.get(self._metric)

        if self._iter[trial_id] < self._grace_period:
            return False

        if self._mode == 'min':
            return True if metric_result > self._metric_threshold else False
        return True if metric_result < self._metric_threshold else False

    def stop_all(self) -> bool:
        """Check if all trials should be stopped.

        Returns:
            bool: Whether to stop all trials.
        """
        return False



# Copyright (c) SI-Analytics. All rights reserved.
import copy
import random
from typing import Callable, Dict, Optional

from ray.tune.experiment import Trial
from ray.tune.schedulers.pbt import \
    PopulationBasedTraining as _PopulationBasedTraining
from ray.tune.search.sample import Domain

from siatune.tune.spaces import build_space
from siatune.utils import ImmutableContainer
from .builder import TRIAL_SCHEDULERS


def explore(
    config: Dict,
    mutations: Dict,
    resample_probability: float,
    custom_explore_fn: Optional[Callable],
) -> Dict:
    """Return a config perturbed as specified.

    Args:
        config: Original hyperparameter configuration.
        mutations: Specification of mutations to perform as documented
            in the PopulationBasedTraining scheduler.
        resample_probability: Probability of allowing resampling of a
            particular variable.
        custom_explore_fn: Custom explore fn applied after built-in
            config perturbations are.
    """
    new_config = copy.deepcopy(config)
    for key, distribution in mutations.items():
        assert isinstance(distribution, Domain)
        if random.random() < resample_probability:
            new_config[key] = ImmutableContainer.decouple(
                distribution.sample(None))

        try:
            new_config[key] = config[key] * 1.2 if random.random(
            ) > 0.5 else config[key] * 0.8
            if isinstance(config[key], int):
                new_config[key] = int(new_config[key])
        except Exception:
            new_config[key] = config[key]
    if custom_explore_fn:
        new_config = custom_explore_fn(new_config)
        assert new_config is not None
    return new_config


@TRIAL_SCHEDULERS.register_module(force=True)
class PopulationBasedTraining(_PopulationBasedTraining):

    def __init__(self, *args, **kwargs) -> None:
        hyperparam_mutations = kwargs.get('hyperparam_mutations',
                                          dict()).copy()
        kwargs.update(hyperparam_mutations=build_space(hyperparam_mutations))
        super().__init__(*args, **kwargs)

    def _get_new_config(self, trial: Trial, trial_to_clone: Trial) -> Trial:
        """Gets new config for trial by exploring trial_to_clone's config.

        Argss:
            trial: The trial to chenge.
            trial_to_clone: The trial to reference.

        Returns:
            Trail: Changed Trial
        """
        return explore(
            trial_to_clone.config,
            self._hyperparam_mutations,
            self._resample_probability,
            self._custom_explore_fn,
        )



# Copyright (c) SI-Analytics. All rights reserved.
from typing import Callable, Optional, Sequence, Union

import ray.tune as tune
from ray.tune.search.sample import Domain

from siatune.utils import ImmutableContainer
from .base import BaseSpace
from .builder import SPACES


@SPACES.register_module()
class Choice(BaseSpace):
    """Sample a categorical value.

    Args:
        categories (Sequence | dict): The categorical search space to choose
            one. If categories is dict, keys of dict will overwrite the alias.
        alias (Sequence, optional): A alias to be expressed. Defaults to None.
    """

    sample: Callable = tune.choice

    def __init__(self,
                 categories: Union[Sequence, dict],
                 alias: Optional[Sequence] = None) -> None:
        if isinstance(categories, dict):
            alias, categories = zip(*[(k, v) for k, v in categories.items()])

        if alias is not None:
            assert isinstance(alias, Sequence)
            assert len(categories) == len(alias)
        alias = alias or [None] * len(categories)
        categories = [ImmutableContainer(*it) for it in zip(categories, alias)]
        self.categories = categories

    @property
    def space(self) -> Domain:
        return self.sample.__func__(self.categories)



# Copyright (c) SI-Analytics. All rights reserved.
from os import path as osp

from mim.utils import get_installed_path, is_installed, module_full_name


def get_train_script(pkg_name: str) -> str:
    """This function returns the path to the train script for a given package.

    Args:
        pkg_name (str): The abbreviation of the package.

    Returns:
        str: The path to the train script.

    Raises:
        ValueError:
            If the given package abbreviation is not unique.
        RuntimeError:
            If the given package is not installed
            or if the train script cannot be found.
    """
    pkg_full_name = module_full_name(pkg_name)
    if pkg_full_name == '':
        raise ValueError("Can't determine a unique "
                         f'package given abbreviation {pkg_name}')
    if not is_installed(pkg_full_name):
        raise RuntimeError(f'The codebase {pkg_name} is not installed')
    pkg_root = get_installed_path(pkg_full_name)
    # tools will be put in package/.mim in PR #68
    train_script = osp.join(pkg_root, '.mim', 'tools', 'train.py')
    if not osp.exists(train_script):
        _alternative = osp.join(pkg_root, 'tools', 'train.py')
        if osp.exists(_alternative):
            train_script = _alternative
        else:
            raise RuntimeError("Can't find train script")
    return train_script



# Copyright (c) SI-Analytics. All rights reserved.
from typing import Callable, Union

import ray.tune as tune
from ray.tune.search.sample import Domain

from .base import BaseSpace
from .builder import SPACES


@SPACES.register_module()
class SampleFrom(BaseSpace):
    """Specify that tune should sample configuration values from this function.

    Args:
        func (str | Callable): An string or callable function
            to draw a sample from.
    """

    sample: Callable = tune.sample_from

    def __init__(self, func: Union[str, Callable]) -> None:
        if isinstance(func, str):
            assert func.startswith('lambda')
            func = eval(func)
        self.func = func

    @property
    def space(self) -> Domain:
        return self.sample.__func__(self.func)



# Copyright (c) SI-Analytics. All rights reserved.
from typing import Callable, Optional, Sequence

import ray.tune as tune

from siatune.utils import ImmutableContainer
from .base import BaseSpace
from .builder import SPACES


@SPACES.register_module()
class GridSearch(BaseSpace):
    """Grid search over a value.

    Args:
        values (Sequence): An iterable whose parameters will be gridded.
        alias (Sequence, optional): A alias to be expressed.
            Defaults to None.
    """

    sample: Callable = tune.grid_search

    def __init__(self,
                 values: Sequence,
                 alias: Optional[Sequence] = None) -> None:
        if alias is not None:
            assert len(values) == len(alias)
            values = [ImmutableContainer(*it) for it in zip(values, alias)]
        self.values = values

    @property
    def space(self) -> dict:
        return self.sample.__func__(self.values)



# Copyright (c) SI-Analytics. All rights reserved.
from typing import Dict

from mmengine.registry import Registry

from .base import BaseTask

TASKS = Registry('task')


def build_task(task: Dict) -> BaseTask:
    """Build the task processor.

    Args:
        task (Dict): The task config to build.

    Returns:
        BaseTask: The task processor.
    """

    return TASKS.build(task)



# Copyright (c) SI-Analytics. All rights reserved.
from typing import Mapping

from mmengine.registry import Registry

SPACES = Registry('space')


def build_space(cfg: dict) -> dict:
    """Build a space.

    Args:
        cfg (dict): The configurations of the space.

    Returns:
        dict: The instantiated space.
    """
    cfg = cfg.copy()
    for k, v in cfg.items():
        if isinstance(v, (int, str, bool, float)):
            continue
        elif isinstance(v, (list, tuple)):
            cfg[k] = type(v)(
                [build_space(_) if isinstance(_, dict) else _ for _ in v])
        elif isinstance(v, Mapping):
            cfg[k] = build_space(v)
            typ = cfg[k].get('type', '')
            if isinstance(typ, str) and typ in SPACES:
                cfg[k] = SPACES.build(cfg[k]).space
    return cfg



# Copyright (c) SI-Analytics. All rights reserved.
import argparse
import copy
import logging
import os
import time
import warnings
from os import path as osp
from typing import Sequence

from mmengine.config import Config, DictAction

from siatune.version import IS_DEPRECATED_MMCV
from .builder import TASKS
from .mm import MMBaseTask

if IS_DEPRECATED_MMCV:

    @TASKS.register_module()
    class MMDetection(MMBaseTask):
        """MMDetection wrapper class for `ray.tune`.

        It is modified from https://github.com/open-mmlab/mmdetection/blob/v2.25.2/tools/train.py

        Args:
            args (argparse.Namespace): The arguments for `tools/train.py`
                script file. It is parsed by :method:`parse_args`.
            num_workers (int): The number of workers to launch.
            num_cpus_per_worker (int): The number of CPUs per worker.
                Default to 1.
            num_gpus_per_worker (int): The number of GPUs per worker.
                Since it must be equal to `num_workers` attribute, it
                is not used in MMDetection.
            rewriters (list[dict] | dict, optional): Context redefinition
                pipeline. Default to None.
        """

        VERSION = 'v2.25.2'

        def parse_args(self, task_args: Sequence[str]):

            parser = argparse.ArgumentParser(description='Train a detector')
            parser.add_argument('config', help='train config file path')
            parser.add_argument(
                '--work-dir', help='the dir to save logs and models')
            parser.add_argument(
                '--resume-from', help='the checkpoint file to resume from')
            parser.add_argument(
                '--auto-resume',
                action='store_true',
                help='resume from the latest checkpoint automatically')
            parser.add_argument(
                '--no-validate',
                action='store_true',
                help='whether not to evaluate the checkpoint during training')
            group_gpus = parser.add_mutually_exclusive_group()
            group_gpus.add_argument(
                '--gpus',
                type=int,
                help='(Deprecated, please use --gpu-id) number of gpus to use '
                '(only applicable to non-distributed training)')
            group_gpus.add_argument(
                '--gpu-ids',
                type=int,
                nargs='+',
                help='(Deprecated, please use --gpu-id) ids of gpus to use '
                '(only applicable to non-distributed training)')
            group_gpus.add_argument(
                '--gpu-id',
                type=int,
                default=0,
                help='id of gpu to use '
                '(only applicable to non-distributed training)')
            parser.add_argument(
                '--seed', type=int, default=None, help='random seed')
            parser.add_argument(
                '--diff-seed',
                action='store_true',
                help='Whether or not set different seeds for different ranks')
            parser.add_argument(
                '--deterministic',
                action='store_true',
                help='whether to set deterministic options for CUDNN backend.')
            parser.add_argument(
                '--options',
                nargs='+',
                action=DictAction,
                help=
                'override some settings in the used config, the key-value pair '
                'in xxx=yyy format will be merged into config file (deprecate), '
                'change to --cfg-options instead.')
            parser.add_argument(
                '--cfg-options',
                nargs='+',
                action=DictAction,
                help=
                'override some settings in the used config, the key-value pair '
                'in xxx=yyy format will be merged into config file. If the value to '
                'be overwritten is a list, it should be like key="[a,b]" or key=a,b '
                'It also allows nested list/tuple values, e.g. key="[(a,b),(c,d)]" '
                'Note that the quotation marks are necessary and that no white space '
                'is allowed.')
            parser.add_argument(
                '--launcher',
                choices=['none', 'pytorch', 'slurm', 'mpi'],
                default='none',
                help='job launcher')
            parser.add_argument('--local_rank', type=int, default=0)
            parser.add_argument(
                '--auto-scale-lr',
                action='store_true',
                help='enable automatically scaling LR.')
            args = parser.parse_args(task_args)
            if 'LOCAL_RANK' not in os.environ:
                os.environ['LOCAL_RANK'] = str(args.local_rank)

            if args.options and args.cfg_options:
                raise ValueError(
                    '--options and --cfg-options cannot be both '
                    'specified, --options is deprecated in favor of --cfg-options'
                )
            if args.options:
                warnings.warn(
                    '--options is deprecated in favor of --cfg-options')
                args.cfg_options = args.options

            return args

        def run(self, args: argparse.Namespace):
            """Run the task.

            Args:
                args (argparse.Namespace):
                    The args that received from context manager.
            """

            import mmcv
            import torch
            import torch.distributed as dist
            from mmcv import Config
            from mmcv.runner import get_dist_info, init_dist
            from mmcv.utils import get_git_hash
            from mmdet import __version__
            from mmdet.apis import (init_random_seed, set_random_seed,
                                    train_detector)
            from mmdet.datasets import build_dataset
            from mmdet.models import build_detector
            from mmdet.utils import (collect_env, get_device, get_root_logger,
                                     replace_cfg_vals, setup_multi_processes,
                                     update_data_root)

            cfg = Config.fromfile(args.config)

            # replace the ${key} with the value of cfg.key
            cfg = replace_cfg_vals(cfg)

            # update data root according to MMDET_DATASETS
            update_data_root(cfg)

            if args.cfg_options is not None:
                cfg.merge_from_dict(args.cfg_options)

            if args.auto_scale_lr:
                if 'auto_scale_lr' in cfg and \
                        'enable' in cfg.auto_scale_lr and \
                        'base_batch_size' in cfg.auto_scale_lr:
                    cfg.auto_scale_lr.enable = True
                else:
                    warnings.warn('Can not find "auto_scale_lr" or '
                                  '"auto_scale_lr.enable" or '
                                  '"auto_scale_lr.base_batch_size" in your'
                                  ' configuration file. Please update all the '
                                  'configuration files to mmdet >= 2.24.1.')

            # set multi-process settings
            setup_multi_processes(cfg)

            # set cudnn_benchmark
            if cfg.get('cudnn_benchmark', False):
                torch.backends.cudnn.benchmark = True

            # work_dir is determined in this priority: CLI > segment in file > filename
            if args.work_dir is not None:
                # update configs according to CLI args if args.work_dir is not None
                cfg.work_dir = args.work_dir
            elif cfg.get('work_dir', None) is None:
                # use config filename as default work_dir if cfg.work_dir is None
                cfg.work_dir = osp.join(
                    './work_dirs',
                    osp.splitext(osp.basename(args.config))[0])

            if args.resume_from is not None:
                cfg.resume_from = args.resume_from
            cfg.auto_resume = args.auto_resume
            if args.gpus is not None:
                cfg.gpu_ids = range(1)
                warnings.warn('`--gpus` is deprecated because we only support '
                              'single GPU mode in non-distributed training. '
                              'Use `gpus=1` now.')
            if args.gpu_ids is not None:
                cfg.gpu_ids = args.gpu_ids[0:1]
                warnings.warn(
                    '`--gpu-ids` is deprecated, please use `--gpu-id`. '
                    'Because we only support single GPU mode in '
                    'non-distributed training. Use the first GPU '
                    'in `gpu_ids` now.')
            if args.gpus is None and args.gpu_ids is None:
                cfg.gpu_ids = [args.gpu_id]

            # init distributed env first, since logger depends on the dist info.
            if args.launcher == 'none':
                distributed = False
            else:
                distributed = True
                init_dist(args.launcher, **cfg.dist_params)
                # re-set gpu_ids with distributed training mode
                _, world_size = get_dist_info()
                cfg.gpu_ids = range(world_size)

            # create work_dir
            mmcv.mkdir_or_exist(osp.abspath(cfg.work_dir))
            # dump config
            cfg.dump(osp.join(cfg.work_dir, osp.basename(args.config)))
            # init the logger before other steps
            timestamp = time.strftime('%Y%m%d_%H%M%S', time.localtime())
            log_file = osp.join(cfg.work_dir, f'{timestamp}.log')
            logger = get_root_logger(
                log_file=log_file, log_level=cfg.log_level)

            # init the meta dict to record some important information such as
            # environment info and seed, which will be logged
            meta = dict()
            # log env info
            env_info_dict = collect_env()
            env_info = '\n'.join([(f'{k}: {v}')
                                  for k, v in env_info_dict.items()])
            dash_line = '-' * 60 + '\n'
            logger.info('Environment info:\n' + dash_line + env_info + '\n' +
                        dash_line)
            meta['env_info'] = env_info
            meta['config'] = cfg.pretty_text
            # log some basic info
            logger.info(f'Distributed training: {distributed}')
            logger.info(f'Config:\n{cfg.pretty_text}')

            cfg.device = get_device()
            # set random seeds
            seed = init_random_seed(args.seed, device=cfg.device)
            seed = seed + dist.get_rank() if args.diff_seed else seed
            logger.info(f'Set random seed to {seed}, '
                        f'deterministic: {args.deterministic}')
            set_random_seed(seed, deterministic=args.deterministic)
            cfg.seed = seed
            meta['seed'] = seed
            meta['exp_name'] = osp.basename(args.config)

            model = build_detector(
                cfg.model,
                train_cfg=cfg.get('train_cfg'),
                test_cfg=cfg.get('test_cfg'))
            model.init_weights()

            datasets = [build_dataset(cfg.data.train)]
            if len(cfg.workflow) == 2:
                assert 'val' in [mode for (mode, _) in cfg.workflow]
                val_dataset = copy.deepcopy(cfg.data.val)
                val_dataset.pipeline = cfg.data.train.get(
                    'pipeline', cfg.data.train.dataset.get('pipeline'))
                datasets.append(build_dataset(val_dataset))
            if cfg.checkpoint_config is not None:
                # save mmdet version, config file content and class names in
                # checkpoints as meta data
                cfg.checkpoint_config.meta = dict(
                    mmdet_version=__version__ + get_git_hash()[:7],
                    CLASSES=datasets[0].CLASSES)
            # add an attribute for visualization convenience
            model.CLASSES = datasets[0].CLASSES
            train_detector(
                model,
                datasets,
                cfg,
                distributed=distributed,
                validate=(not args.no_validate),
                timestamp=timestamp,
                meta=meta)

else:

    @TASKS.register_module()
    class MMDetection(MMBaseTask):
        """MMDetection wrapper class for `ray.tune`.

        It is modified from https://github.com/open-mmlab/mmdetection/blob/v3.0.0rc4/tools/train.py

        Args:
            args (argparse.Namespace): The arguments for `tools/train.py`
                script file. It is parsed by :method:`parse_args`.
            num_workers (int): The number of workers to launch.
            num_cpus_per_worker (int): The number of CPUs per worker.
                Default to 1.
            num_gpus_per_worker (int): The number of GPUs per worker.
                Since it must be equal to `num_workers` attribute, it
                is not used in MMDetection.
            rewriters (list[dict] | dict, optional): Context redefinition
                pipeline. Default to None.
        """

        VERSION = 'v3.0.0rc4'

        def parse_args(self, task_args: Sequence[str]):

            parser = argparse.ArgumentParser(description='Train a detector')
            parser.add_argument('config', help='train config file path')
            parser.add_argument(
                '--work-dir', help='the dir to save logs and models')
            parser.add_argument(
                '--amp',
                action='store_true',
                default=False,
                help='enable automatic-mixed-precision training')
            parser.add_argument(
                '--auto-scale-lr',
                action='store_true',
                help='enable automatically scaling LR.')
            parser.add_argument(
                '--resume',
                nargs='?',
                type=str,
                const='auto',
                help='If specify checkpoint path, resume from it, while if not '
                'specify, try to auto resume from the latest checkpoint '
                'in the work directory.')
            parser.add_argument(
                '--cfg-options',
                nargs='+',
                action=DictAction,
                help=
                'override some settings in the used config, the key-value pair '
                'in xxx=yyy format will be merged into config file. If the value to '
                'be overwritten is a list, it should be like key="[a,b]" or key=a,b '
                'It also allows nested list/tuple values, e.g. key="[(a,b),(c,d)]" '
                'Note that the quotation marks are necessary and that no white space '
                'is allowed.')
            parser.add_argument(
                '--launcher',
                choices=['none', 'pytorch', 'slurm', 'mpi'],
                default='none',
                help='job launcher')
            parser.add_argument('--local_rank', type=int, default=0)
            args = parser.parse_args(task_args)
            if 'LOCAL_RANK' not in os.environ:
                os.environ['LOCAL_RANK'] = str(args.local_rank)

            return args

        def run(self, args: argparse.Namespace):
            """Run the task.

            Args:
                args (argparse.Namespace):
                    The args that received from context manager.
            """
            from mmdet.utils import register_all_modules
            from mmengine.logging import print_log
            from mmengine.runner import RUNNERS, Runner

            # register all modules in mmdet into the registries
            # do not init the default scope here because it will be init in the runner
            register_all_modules(init_default_scope=False)

            # load config
            cfg = Config.fromfile(args.config)
            cfg.launcher = args.launcher
            if args.cfg_options is not None:
                cfg.merge_from_dict(args.cfg_options)

            # work_dir is determined in this priority: CLI > segment in file > filename
            if args.work_dir is not None:
                # update configs according to CLI args if args.work_dir is not None
                cfg.work_dir = args.work_dir
            elif cfg.get('work_dir', None) is None:
                # use config filename as default work_dir if cfg.work_dir is None
                cfg.work_dir = osp.join(
                    './work_dirs',
                    osp.splitext(osp.basename(args.config))[0])

            # enable automatic-mixed-precision training
            if args.amp is True:
                optim_wrapper = cfg.optim_wrapper.type
                if optim_wrapper == 'AmpOptimWrapper':
                    print_log(
                        'AMP training is already enabled in your config.',
                        logger='current',
                        level=logging.WARNING)
                else:
                    assert optim_wrapper == 'OptimWrapper', (
                        '`--amp` is only supported when the optimizer wrapper type is '
                        f'`OptimWrapper` but got {optim_wrapper}.')
                    cfg.optim_wrapper.type = 'AmpOptimWrapper'
                    cfg.optim_wrapper.loss_scale = 'dynamic'

            # enable automatically scaling LR
            if args.auto_scale_lr:
                if 'auto_scale_lr' in cfg and \
                        'enable' in cfg.auto_scale_lr and \
                        'base_batch_size' in cfg.auto_scale_lr:
                    cfg.auto_scale_lr.enable = True
                else:
                    raise RuntimeError(
                        'Can not find "auto_scale_lr" or '
                        '"auto_scale_lr.enable" or '
                        '"auto_scale_lr.base_batch_size" in your'
                        ' configuration file.')

            # resume is determined in this priority: resume from > auto_resume
            if args.resume == 'auto':
                cfg.resume = True
                cfg.load_from = None
            elif args.resume is not None:
                cfg.resume = True
                cfg.load_from = args.resume

            # build the runner from config
            if 'runner_type' not in cfg:
                # build the default runner
                runner = Runner.from_cfg(cfg)
            else:
                # build customized runner from the registry
                # if 'runner_type' is set in the cfg
                runner = RUNNERS.build(cfg)

            # start training
            runner.train()



# Copyright (c) SI-Analytics. All rights reserved.
# Modified from https://github.com/ray-project/ray/blob/ray-2.1.0/python/ray/train/torch/config.py  # noqa

import logging
import os
from dataclasses import dataclass

import ray
import torch.distributed as dist
from ray.train._internal.utils import get_address_and_port
from ray.train._internal.worker_group import WorkerGroup
from ray.train.backend import BackendConfig
from ray.train.torch.config import _set_nccl_network_interface, _TorchBackend

logger = logging.getLogger(__name__)


@dataclass
class MMBackendConfig(BackendConfig):
    """Configuration for torch process group setup."""

    @property
    def backend_cls(self):
        return _CustomTorchBackend


class _CustomTorchBackend(_TorchBackend):

    def on_start(self, worker_group: WorkerGroup,
                 backend_config: BackendConfig):
        if not dist.is_available():
            raise RuntimeError('Distributed torch is not available.')

        worker_group.execute(_set_nccl_network_interface)

        master_addr, master_port = worker_group.execute_single(
            0, get_address_and_port)

        def set_env_vars(addr, port, rank, world_size):
            os.environ['MASTER_ADDR'] = addr
            os.environ['MASTER_PORT'] = str(port)
            os.environ['RANK'] = str(rank)
            os.environ['LOCAL_RANK'] = str(rank)
            os.environ['WORLD_SIZE'] = str(world_size)

        setup_futures = []
        for i in range(len(worker_group)):
            setup_futures.append(
                worker_group.execute_single_async(
                    i,
                    set_env_vars,
                    addr=master_addr,
                    port=master_port,
                    rank=i,
                    world_size=len(worker_group),
                ))
        ray.get(setup_futures)



# Copyright (c) SI-Analytics. All rights reserved.
from importlib.machinery import SourceFileLoader
from typing import Sequence

from siatune.utils import get_train_script
from .builder import TASKS
from .mm import MMBaseTask


class _EntrypointExecutor:
    """Execute the entrypoint of open mm train-based projects.

    Args:
        pkg_name (str): The abbreviation of the package.
        argv (Sequence[str]): The arguments for `tools/train.py`
        module_name (str):
            The name of the module to execute. Defaults to 'main'.
    """

    def __init__(self,
                 pkg_name: str,
                 argv: Sequence[str],
                 module_name: str = 'main'):
        self._train_script = get_train_script(pkg_name)
        self._module_name = module_name
        self._argv = argv
        self._entrypoint = SourceFileLoader(self._module_name,
                                            self._train_script).load_module()

    def _hijack_argv(self, argv: Sequence[str]):
        """Hijack the command line arguments.

        Args:
            argv (Sequence[str]): The arguments for `tools/train.py`
        """
        import sys
        sys.argv[1:] = argv
        return

    def execute(self):
        """Run the task."""
        self._hijack_argv(self._argv)
        getattr(self._entrypoint, self._module_name)()


@TASKS.register_module()
class MIM(MMBaseTask):
    """Wrapper class execute any script provided by all OpenMMLab codebases.

    Args:
        pkg_name (str): The abbreviation of the package.
    """

    def __init__(self, pkg_name: str, **kwargs):
        self._pkg_name = pkg_name
        super().__init__(should_parse=False, **kwargs)

    def parse_args(self, *args, **kwargs) -> None:
        pass

    def run(self, args: Sequence[str]):
        """This method runs a task in the MIM framework.

        Args:
            args (Sequence[str]): A list of command-line arguments.
        """
        executor = _EntrypointExecutor(self._pkg_name, args)
        executor.execute()



# Copyright (c) SI-Analytics. All rights reserved.
import copy
import os.path as osp
import time
from typing import Callable, Optional, Union

import mmengine
from ray.air.config import RunConfig
from ray.tune import ResultGrid
from ray.tune.tune_config import TuneConfig
from ray.tune.tuner import Tuner as RayTuner

from siatune.codebase import build_task
from siatune.tune import (build_callback, build_scheduler, build_searcher,
                          build_space, build_stopper)
from .utils import NAME_CREATOR


class Tuner:
    """Wrapper class of :class:`ray.tune.tuner.Tuner`.

    Args:
        task (dict | Callable): The trainable task to be tuned.
        work_dir (str): The working directory to save checkpoints. The logs
            will be saved in the subdirectory of `work_dir`.
        param_space (dict, optional): Search space of the tuning task.
        tune_cfg (dict, optional): Tuning algorithm specific configs
            except for `search_alg` and `scheduler`.
            Refer to :class:`ray.tune.tune_config.TuneConfig` for more info.
        searcher (dict, optional): Search algorithm for optimization.
            Default to random search.
            Refer to :module:`ray.tune.search` for more options.
        trial_scheduler (dict, optional): Scheduler for executing the trial.
            Default to FIFO scheduler.
            Refer to :module:`ray.tune.schedulers` for more options.
        stopper (dict, optional): Stop conditions to consider.
            Refer to :class:`ray.tune.stopper.Stopper` for more info.
        callbacks (dict | list, optional): Callbacks to invoke.
            Refer to :class:`ray.tune.callback.Callback` for more info.
        resume (str, optional): The experiment path to resume.
            Default to None.
        experiment_name (str, optional): Name of current experiment. If not
            specified, timestamp will be used as `experiment_name`.
            Default to None.
        cfg (dict, optional) Full config. Default to None.
    """

    def __init__(
        self,
        task: Union[dict, Callable[[dict], None]],
        work_dir: str,
        param_space: Optional[dict] = None,
        tune_cfg: Optional[dict] = None,
        searcher: Optional[dict] = None,
        trial_scheduler: Optional[dict] = None,
        stopper: Optional[dict] = None,
        callbacks: Optional[Union[dict, list]] = None,
        resume: Optional[str] = None,
        experiment_name: Optional[str] = None,
        cfg: Optional[dict] = None,
    ):
        trainable = task
        if isinstance(task, dict):
            task = build_task(task)
            trainable = task.create_trainable()

        self.work_dir = osp.abspath(work_dir)
        mmengine.mkdir_or_exist(self.work_dir)

        if param_space is not None:
            param_space = build_space(param_space)

        tune_cfg = copy.deepcopy(tune_cfg or dict())

        if searcher is not None:
            searcher = build_searcher(searcher)

        if trial_scheduler is not None:
            trial_scheduler = build_scheduler(trial_scheduler)

        if stopper is not None:
            stopper = build_stopper(stopper)

        if callbacks is not None:
            if isinstance(callbacks, dict):
                callbacks = [callbacks]
            callbacks = [build_callback(callback) for callback in callbacks]

        self.resume = resume

        if experiment_name is None:
            experiment_name = time.strftime('%Y%m%d_%H%M%S', time.localtime())
        self.experiment_name = experiment_name

        self.cfg = cfg

        # dump `cfg` to `work_dir`
        filename = self.cfg.filename or f'{experiment_name}.py'
        if self.cfg.filename is not None:
            filename = osp.basename(self.cfg.filename)
        mmengine.mkdir_or_exist(osp.join(self.work_dir, experiment_name))
        self.cfg.dump(osp.join(self.work_dir, experiment_name, filename))

        self.tuner = RayTuner(
            trainable,
            param_space=param_space,
            tune_config=TuneConfig(
                search_alg=searcher,
                scheduler=trial_scheduler,
                trial_name_creator=NAME_CREATOR.get(
                    tune_cfg.pop('trial_name_creator', 'trial_id')),
                trial_dirname_creator=NAME_CREATOR.get(
                    tune_cfg.pop('trial_dirname_creator', 'experiment_tag')),
                **tune_cfg),
            run_config=RunConfig(
                name=self.experiment_name,
                local_dir=self.work_dir,
                stop=stopper,
                callbacks=callbacks,
                failure_config=None,  # todo
                sync_config=None,  # todo
                checkpoint_config=None,  # todo
            ),
        )

    @classmethod
    def from_cfg(cls, cfg: dict) -> 'Tuner':
        cfg = copy.deepcopy(cfg)
        tuner = cls(
            task=cfg['task'],
            work_dir=cfg['work_dir'],
            param_space=cfg.get('space', None),
            tune_cfg=cfg.get('tune_cfg', None),
            searcher=cfg.get('searcher', None),
            trial_scheduler=cfg.get('trial_scheduler', None),
            stopper=cfg.get('stopper', None),
            callbacks=cfg.get('callbacks', None),
            resume=cfg.get('resume', None),
            experiment_name=cfg.get('experiment_name', None),
            cfg=cfg,
        )

        return tuner

    def tune(self) -> ResultGrid:
        """Launch tuning.

        Returns:
            ResultGrid: The :class:`ResultGrid` instance after tuning.
        """
        if self.resume is not None:
            self.tuner = RayTuner.restore(self.resume)

        return self.tuner.fit()



# Copyright (c) SI-Analytics. All rights reserved.
import argparse
from abc import ABCMeta, abstractmethod
from copy import deepcopy
from typing import Optional, Sequence, Union

from ray.tune import Trainable

from siatune.core import ContextManager
from siatune.utils import ImmutableContainer


class BaseTask(metaclass=ABCMeta):
    """Base class to specify the target task.

    The main functions of the task processor are as follows:
    1. CLI argument definition and parsing. (`parse_args`)
        Inputs: args (Sequence[str])
        Outputs: args (argparse.Namespace)
    2. Objective function execution.
        The result must be reported by calling tune.report. (`run`)
        Inputs: args (argparse.Namespace)
        Outputs: None
    3. Gather and refine information from multiple sources.
        Call the 'run' function of the task processor. (`context_aware_run`)
        Aggregate the information we define as context,
        convert it into a refined argparse namespace, and input it to run.
        The context consists of:
            1. args (argparse.Namespace | Sequence[str]):
                The low level CLI arguments.
            2. searched_cfg (Dict):
                The configuration searched by the algorithm.
        Inputs: searched_cfg (Dict)
        Outputs: None

    Args:
        args (Sequence[str]): The task arguments. It is parsed by
            :method:`parse_args`.
        num_workers (int): The number of workers to launch.
        num_cpus_per_worker (int): The number of CPUs per worker.
            Default to 1.
        num_gpus_per_worker (int): The number of GPUs per worker.
            Default to 1.
        rewriters (list[dict] | dict, optional): Context redefinition
            pipeline. Default to None.
    """

    def __init__(self,
                 args: Sequence[str],
                 num_workers: int = 1,
                 num_cpus_per_worker: int = 1,
                 num_gpus_per_worker: int = 1,
                 rewriters: Optional[Union[list, dict]] = None,
                 should_parse: bool = True):

        if should_parse:
            args = self.parse_args(args)
        self.args = args

        self.num_workers = num_workers
        self.num_cpus_per_worker = num_cpus_per_worker
        self.num_gpus_per_worker = num_gpus_per_worker

        if isinstance(rewriters, dict):
            rewriters = [rewriters]
        self.rewriters = rewriters

    @abstractmethod
    def parse_args(self,
                   args: Sequence[str]) -> Union[argparse.Namespace, None]:
        """Define and parse the necessary arguments for the task.

        Args:
            args (Sequence[str]): The args.

        Returns:
            argparse.Namespace: The parsed args.
        """
        pass

    def context_aware_run(self, searched_cfg: dict):
        """Gather and refine the information received by users and Ray.tune to
        execute the objective task.

        Args:
            searched_cfg (Dict): The searched configuration.
        """

        context_manager = ContextManager(self.rewriters)
        context = dict(
            args=deepcopy(self.args),
            searched_cfg=deepcopy(ImmutableContainer.decouple(searched_cfg)),
        )
        return context_manager(self.run)(**context)

    @abstractmethod
    def run(self, *args, **kwargs):
        """Run the trainable task."""
        pass

    @abstractmethod
    def create_trainable(self) -> Trainable:
        """Get ray trainable task."""
        pass



# Copyright (c) SI-Analytics. All rights reserved.
from .args import reference_raw_args
from .config import dump_cfg
from .container import ImmutableContainer
from .dist import set_env_vars
from .logger import get_root_logger
from .mim import get_train_script
from .setup_env import register_all_modules

__all__ = [
    'ImmutableContainer', 'dump_cfg', 'get_root_logger', 'reference_raw_args',
    'get_train_script', 'set_env_vars', 'register_all_modules'
]



# Copyright (c) SI-Analytics. All rights reserved.
import argparse
import copy
import logging
import os
import time
import warnings
from os import path as osp
from typing import Sequence

from mmengine.config import Config, DictAction

from siatune.version import IS_DEPRECATED_MMCV
from .builder import TASKS
from .mm import MMBaseTask

if IS_DEPRECATED_MMCV:

    @TASKS.register_module()
    class MMSegmentation(MMBaseTask):
        """MMSegmentation wrapper class for `ray.tune`.

        It is modified from https://github.com/open-mmlab/mmsegmentation/blob/v0.25.0/tools/train.py

        Args:
            args (argparse.Namespace): The arguments for `tools/train.py`
                script file. It is parsed by :method:`parse_args`.
            num_workers (int): The number of workers to launch.
            num_cpus_per_worker (int): The number of CPUs per worker.
                Default to 1.
            num_gpus_per_worker (int): The number of GPUs per worker.
                Since it must be equal to `num_workers` attribute, it
                is not used in MMSegmentation.
            rewriters (list[dict] | dict, optional): Context redefinition
                pipeline. Default to None.
        """

        VERSION = 'v0.25.0'

        def parse_args(self, task_args: Sequence[str]):
            parser = argparse.ArgumentParser(description='Train a segmentor')
            parser.add_argument('config', help='train config file path')
            parser.add_argument(
                '--work-dir', help='the dir to save logs and models')
            parser.add_argument(
                '--load-from', help='the checkpoint file to load weights from')
            parser.add_argument(
                '--resume-from', help='the checkpoint file to resume from')
            parser.add_argument(
                '--no-validate',
                action='store_true',
                help='whether not to evaluate the checkpoint during training')
            group_gpus = parser.add_mutually_exclusive_group()
            group_gpus.add_argument(
                '--gpus',
                type=int,
                help='(Deprecated, please use --gpu-id) number of gpus to use '
                '(only applicable to non-distributed training)')
            group_gpus.add_argument(
                '--gpu-ids',
                type=int,
                nargs='+',
                help='(Deprecated, please use --gpu-id) ids of gpus to use '
                '(only applicable to non-distributed training)')
            group_gpus.add_argument(
                '--gpu-id',
                type=int,
                default=0,
                help='id of gpu to use '
                '(only applicable to non-distributed training)')
            parser.add_argument(
                '--seed', type=int, default=None, help='random seed')
            parser.add_argument(
                '--diff_seed',
                action='store_true',
                help='Whether or not set different seeds for different ranks')
            parser.add_argument(
                '--deterministic',
                action='store_true',
                help='whether to set deterministic options for CUDNN backend.')
            parser.add_argument(
                '--options',
                nargs='+',
                action=DictAction,
                help=
                "--options is deprecated in favor of --cfg_options' and it will "
                'not be supported in version v0.22.0. Override some settings in the '
                'used config, the key-value pair in xxx=yyy format will be merged '
                'into config file. If the value to be overwritten is a list, it '
                'should be like key="[a,b]" or key=a,b It also allows nested '
                'list/tuple values, e.g. key="[(a,b),(c,d)]" Note that the quotation '
                'marks are necessary and that no white space is allowed.')
            parser.add_argument(
                '--cfg-options',
                nargs='+',
                action=DictAction,
                help=
                'override some settings in the used config, the key-value pair '
                'in xxx=yyy format will be merged into config file. If the value to '
                'be overwritten is a list, it should be like key="[a,b]" or key=a,b '
                'It also allows nested list/tuple values, e.g. key="[(a,b),(c,d)]" '
                'Note that the quotation marks are necessary and that no white space '
                'is allowed.')
            parser.add_argument(
                '--launcher',
                choices=['none', 'pytorch', 'slurm', 'mpi'],
                default='none',
                help='job launcher')
            parser.add_argument('--local_rank', type=int, default=0)
            parser.add_argument(
                '--auto-resume',
                action='store_true',
                help='resume from the latest checkpoint automatically.')
            args = parser.parse_args(task_args)
            if 'LOCAL_RANK' not in os.environ:
                os.environ['LOCAL_RANK'] = str(args.local_rank)

            if args.options and args.cfg_options:
                raise ValueError(
                    '--options and --cfg-options cannot be both '
                    'specified, --options is deprecated in favor of --cfg-options. '
                    '--options will not be supported in version v0.22.0.')
            if args.options:
                warnings.warn(
                    '--options is deprecated in favor of --cfg-options. '
                    '--options will not be supported in version v0.22.0.')
                args.cfg_options = args.options

            return args

        def run(self, args: argparse.Namespace):
            """Run the task.

            Args:
                args (argparse.Namespace):
                    The args that received from context manager.
            """
            import mmcv
            import torch
            import torch.distributed as dist
            from mmcv.cnn.utils import revert_sync_batchnorm
            from mmcv.runner import get_dist_info, init_dist
            from mmcv.utils import Config, get_git_hash
            from mmseg import __version__
            from mmseg.apis import (init_random_seed, set_random_seed,
                                    train_segmentor)
            from mmseg.datasets import build_dataset
            from mmseg.models import build_segmentor
            from mmseg.utils import (collect_env, get_device, get_root_logger,
                                     setup_multi_processes)

            cfg = Config.fromfile(args.config)
            if args.cfg_options is not None:
                cfg.merge_from_dict(args.cfg_options)

            # set cudnn_benchmark
            if cfg.get('cudnn_benchmark', False):
                torch.backends.cudnn.benchmark = True

            # work_dir is determined in this priority: CLI > segment in file > filename
            if args.work_dir is not None:
                # update configs according to CLI args if args.work_dir is not None
                cfg.work_dir = args.work_dir
            elif cfg.get('work_dir', None) is None:
                # use config filename as default work_dir if cfg.work_dir is None
                cfg.work_dir = osp.join(
                    './work_dirs',
                    osp.splitext(osp.basename(args.config))[0])
            if args.load_from is not None:
                cfg.load_from = args.load_from
            if args.resume_from is not None:
                cfg.resume_from = args.resume_from
            if args.gpus is not None:
                cfg.gpu_ids = range(1)
                warnings.warn('`--gpus` is deprecated because we only support '
                              'single GPU mode in non-distributed training. '
                              'Use `gpus=1` now.')
            if args.gpu_ids is not None:
                cfg.gpu_ids = args.gpu_ids[0:1]
                warnings.warn(
                    '`--gpu-ids` is deprecated, please use `--gpu-id`. '
                    'Because we only support single GPU mode in '
                    'non-distributed training. Use the first GPU '
                    'in `gpu_ids` now.')
            if args.gpus is None and args.gpu_ids is None:
                cfg.gpu_ids = [args.gpu_id]

            cfg.auto_resume = args.auto_resume

            # init distributed env first, since logger depends on the dist info.
            if args.launcher == 'none':
                distributed = False
            else:
                distributed = True
                init_dist(args.launcher, **cfg.dist_params)
                # gpu_ids is used to calculate iter when resuming checkpoint
                _, world_size = get_dist_info()
                cfg.gpu_ids = range(world_size)

            # create work_dir
            mmcv.mkdir_or_exist(osp.abspath(cfg.work_dir))
            # dump config
            cfg.dump(osp.join(cfg.work_dir, osp.basename(args.config)))
            # init the logger before other steps
            timestamp = time.strftime('%Y%m%d_%H%M%S', time.localtime())
            log_file = osp.join(cfg.work_dir, f'{timestamp}.log')
            logger = get_root_logger(
                log_file=log_file, log_level=cfg.log_level)

            # set multi-process settings
            setup_multi_processes(cfg)

            # init the meta dict to record some important information such as
            # environment info and seed, which will be logged
            meta = dict()
            # log env info
            env_info_dict = collect_env()
            env_info = '\n'.join(
                [f'{k}: {v}' for k, v in env_info_dict.items()])
            dash_line = '-' * 60 + '\n'
            logger.info('Environment info:\n' + dash_line + env_info + '\n' +
                        dash_line)
            meta['env_info'] = env_info

            # log some basic info
            logger.info(f'Distributed training: {distributed}')
            logger.info(f'Config:\n{cfg.pretty_text}')

            # set random seeds
            cfg.device = get_device()
            seed = init_random_seed(args.seed, device=cfg.device)
            seed = seed + dist.get_rank() if args.diff_seed else seed
            logger.info(f'Set random seed to {seed}, '
                        f'deterministic: {args.deterministic}')
            set_random_seed(seed, deterministic=args.deterministic)
            cfg.seed = seed
            meta['seed'] = seed
            meta['exp_name'] = osp.basename(args.config)

            model = build_segmentor(
                cfg.model,
                train_cfg=cfg.get('train_cfg'),
                test_cfg=cfg.get('test_cfg'))
            model.init_weights()

            # SyncBN is not support for DP
            if not distributed:
                warnings.warn(
                    'SyncBN is only supported with DDP. To be compatible with DP, '
                    'we convert SyncBN to BN. Please use dist_train.sh which can '
                    'avoid this error.')
                model = revert_sync_batchnorm(model)

            logger.info(model)

            datasets = [build_dataset(cfg.data.train)]
            if len(cfg.workflow) == 2:
                val_dataset = copy.deepcopy(cfg.data.val)
                val_dataset.pipeline = cfg.data.train.pipeline
                datasets.append(build_dataset(val_dataset))
            if cfg.checkpoint_config is not None:
                # save mmseg version, config file content and class names in
                # checkpoints as meta data
                cfg.checkpoint_config.meta = dict(
                    mmseg_version=f'{__version__}+{get_git_hash()[:7]}',
                    config=cfg.pretty_text,
                    CLASSES=datasets[0].CLASSES,
                    PALETTE=datasets[0].PALETTE)
            # add an attribute for visualization convenience
            model.CLASSES = datasets[0].CLASSES
            # passing checkpoint meta for saving best checkpoint
            meta.update(cfg.checkpoint_config.meta)
            train_segmentor(
                model,
                datasets,
                cfg,
                distributed=distributed,
                validate=(not args.no_validate),
                timestamp=timestamp,
                meta=meta)

else:

    @TASKS.register_module()
    class MMSegmentation(MMBaseTask):
        """MMSegmentation wrapper class for `ray.tune`.

        It is modified from https://github.com/open-mmlab/mmsegmentation/tree/v1.0.0rc2/tools/train.py

        Args:
            args (argparse.Namespace): The arguments for `tools/train.py`
                script file. It is parsed by :method:`parse_args`.
            num_workers (int): The number of workers to launch.
            num_cpus_per_worker (int): The number of CPUs per worker.
                Default to 1.
            num_gpus_per_worker (int): The number of GPUs per worker.
                Since it must be equal to `num_workers` attribute, it
                is not used in MMSegmentation.
            rewriters (list[dict] | dict, optional): Context redefinition
                pipeline. Default to None.
        """

        VERSION = 'v1.0.0rc2'

        def parse_args(self, task_args: Sequence[str]):
            parser = argparse.ArgumentParser(description='Train a segmentor')
            parser.add_argument('config', help='train config file path')
            parser.add_argument(
                '--work-dir', help='the dir to save logs and models')
            parser.add_argument(
                '--resume',
                action='store_true',
                default=False,
                help=
                'resume from the latest checkpoint in the work_dir automatically'
            )
            parser.add_argument(
                '--amp',
                action='store_true',
                default=False,
                help='enable automatic-mixed-precision training')
            parser.add_argument(
                '--cfg-options',
                nargs='+',
                action=DictAction,
                help=
                'override some settings in the used config, the key-value pair '
                'in xxx=yyy format will be merged into config file. If the value to '
                'be overwritten is a list, it should be like key="[a,b]" or key=a,b '
                'It also allows nested list/tuple values, e.g. key="[(a,b),(c,d)]" '
                'Note that the quotation marks are necessary and that no white space '
                'is allowed.')
            parser.add_argument(
                '--launcher',
                choices=['none', 'pytorch', 'slurm', 'mpi'],
                default='none',
                help='job launcher')
            parser.add_argument('--local_rank', type=int, default=0)
            args = parser.parse_args(task_args)
            if 'LOCAL_RANK' not in os.environ:
                os.environ['LOCAL_RANK'] = str(args.local_rank)

            return args

        def run(self, args: argparse.Namespace):
            """Run the task.

            Args:
                args (argparse.Namespace):
                    The args that received from context manager.
            """
            from mmengine.logging import print_log
            from mmengine.registry import RUNNERS
            from mmengine.runner import Runner
            from mmseg.utils import register_all_modules

            # register all modules in mmseg into the registries
            # do not init the default scope here because it will be init in the runner
            register_all_modules(init_default_scope=False)

            # load config
            cfg = Config.fromfile(args.config)
            cfg.launcher = args.launcher
            if args.cfg_options is not None:
                cfg.merge_from_dict(args.cfg_options)

            # work_dir is determined in this priority: CLI > segment in file > filename
            if args.work_dir is not None:
                # update configs according to CLI args if args.work_dir is not None
                cfg.work_dir = args.work_dir
            elif cfg.get('work_dir', None) is None:
                # use config filename as default work_dir if cfg.work_dir is None
                cfg.work_dir = osp.join(
                    './work_dirs',
                    osp.splitext(osp.basename(args.config))[0])

            # enable automatic-mixed-precision training
            if args.amp is True:
                optim_wrapper = cfg.optim_wrapper.type
                if optim_wrapper == 'AmpOptimWrapper':
                    print_log(
                        'AMP training is already enabled in your config.',
                        logger='current',
                        level=logging.WARNING)
                else:
                    assert optim_wrapper == 'OptimWrapper', (
                        '`--amp` is only supported when the optimizer wrapper type is '
                        f'`OptimWrapper` but got {optim_wrapper}.')
                    cfg.optim_wrapper.type = 'AmpOptimWrapper'
                    cfg.optim_wrapper.loss_scale = 'dynamic'

            # resume training
            cfg.resume = args.resume

            # build the runner from config
            if 'runner_type' not in cfg:
                # build the default runner
                runner = Runner.from_cfg(cfg)
            else:
                # build customized runner from the registry
                # if 'runner_type' is set in the cfg
                runner = RUNNERS.build(cfg)

            # start training
            runner.train()



# Copyright (c) SI-Analytics. All rights reserved.
import logging
from typing import Optional

from mmengine.logging import MMLogger


def get_root_logger(log_file: Optional[str] = None,
                    log_level: int = logging.INFO):
    """Get the root logger. The logger will be initialized if it has not been
    initialized. By default a StreamHandler will be added. If `log_file` is
    specified, a FileHandler will also be added. The name of the root logger is
    the top-level package name, e.g., "siatune".

    Args:
        log_file (Optional[str]): The log filename. If specified, a FileHandler
            will be added to the root logger.
        log_level (int): The root logger level. Note that only the process of
            rank 0 is affected, while other processes will set the level to
            "Error" and be silent most of the time.
    Returns:
        logging.Logger: The root logger.
    """

    logger = MMLogger(name='SIATune', log_file=log_file, log_level=log_level)

    return logger



# Copyright (c) SI-Analytics. All rights reserved.
from typing import Dict, List

from .base import BaseRewriter
from .builder import REWRITERS


@REWRITERS.register_module()
class CustomHookRegister(BaseRewriter):
    """Register custom hooks."""

    def __init__(self,
                 key: str,
                 post_custom_hooks: List[Dict],
                 hk_key: str = 'custom_hooks') -> None:
        """Initialize the rewriter.

        Args:
            key (str): The key of the context.
            post_custom_hooks (List[Dict]): The custom hooks to be registered.
            hk_key (str): The key of the custom hooks.
        """
        # Re-register the hook in the new process.
        from siatune.core import hooks  # noqa F401

        self.post_custom_hooks = post_custom_hooks
        self.key = key
        self.hk_key = hk_key

    def __call__(self, context: Dict) -> Dict:
        """Register custom hooks.

        Args:
            context (Dict): The context to be rewritten.

        Returns:
            Dict: The context after rewriting.
        """

        custom_hooks = getattr(context[self.key], self.hk_key, []).copy()
        for custom_hook in self.post_custom_hooks:
            custom_hooks.append(custom_hook)
        setattr(context[self.key], self.hk_key, custom_hooks)
        return context



# Copyright (c) SI-Analytics. All rights reserved.
import argparse
from typing import Dict

from ray.air import session

from siatune.utils import reference_raw_args
from .base import BaseRewriter
from .builder import REWRITERS


@REWRITERS.register_module()
class ResumeFromCkpt(BaseRewriter):
    """Specifies the checkpoint for resuming training."""
    arg_name: str = 'resume_from'
    raw_arg_name: str = '--resume-from'

    def __call__(self, context: Dict) -> Dict:
        """Set with checkpoints specified by Ray.

        Args:
            context (Dict): The context to be rewritten.
        Returns:
            Dict: The context after rewriting.
        """
        ckpt = session.get_checkpoint()
        if not ckpt:
            return context
        ckpt_path = ckpt.to_dict().get('path')
        is_parsed = isinstance(context['args'], argparse.Namespace)
        if is_parsed:
            setattr(context['args'], self.arg_name, ckpt_path)
        else:
            _, idx = reference_raw_args(context['args'], self.raw_arg_name)
            assert len(idx) < 2
            if idx:
                context['args'][idx.pop()] = ckpt_path
            else:
                context['args'].extend([self.raw_arg_name, ckpt_path])
        return context



# Copyright (c) SI-Analytics. All rights reserved.
from .base import BaseTask
from .blackbox import BlackBoxTask
from .builder import TASKS, build_task
from .cont_test_func import ContinuousTestFunction
from .disc_test_func import DiscreteTestFunction
from .mim import MIM
from .mm import MMBaseTask
from .mmcls import MMClassification
from .mmdet import MMDetection
from .mmedit import MMEditing
from .mmseg import MMSegmentation

__all__ = [
    'TASKS', 'build_task', 'BaseTask', 'BlackBoxTask',
    'ContinuousTestFunction', 'DiscreteTestFunction', 'MMBaseTask',
    'MMClassification', 'MMDetection', 'MMEditing', 'MMSegmentation', 'MIM'
]



# Copyright (c) SI-Analytics. All rights reserved.
from typing import List, Tuple


def reference_raw_args(raw_args: List[str], key: str) -> Tuple:
    """Extract the list of arguments and their indices from a list of command-
    line arguments.

    The list of arguments is considered
    to start after the first occurrence of the given key
    and to end at the next occurrence
    of a key (a string starting with '--').

    Args:
        raw_args (List[str]): List of command-line arguments.
        key (str):
            Key to look for in the raw arguments.
            The key should start with '--'.

    Returns:
        Tuple[List[str], List[int]]: The list of arguments and their indices.
    """
    assert key.startswith('--')
    ret: List[str] = []
    ret_indices: List[int] = []
    if key not in raw_args:
        return ret, ret_indices
    for idx in range(raw_args.index(key) + 1, len(raw_args)):
        cand = raw_args[idx]
        if cand.startswith('--'):
            break
        ret.append(cand)
        ret_indices.append(idx)
    return ret, ret_indices



# Copyright (c) SI-Analytics. All rights reserved.
from mmengine.config import Config
from mmengine.registry import Registry
from ray.tune.logger import (CSVLoggerCallback, JsonLoggerCallback,
                             LegacyLoggerCallback, LoggerCallback,
                             TBXLoggerCallback)

CALLBACKS = Registry('callback')
CALLBACKS.register_module(module=LegacyLoggerCallback)
CALLBACKS.register_module(module=JsonLoggerCallback)
CALLBACKS.register_module(module=CSVLoggerCallback)
CALLBACKS.register_module(module=TBXLoggerCallback)


def build_callback(cfg: Config) -> LoggerCallback:
    return CALLBACKS.build(cfg)



# Copyright (c) SI-Analytics. All rights reserved.
import glob
import os
import os.path as osp
from pathlib import Path
from typing import Dict, Optional, Union

from ray.air import session
from ray.air.checkpoint import Checkpoint

from siatune.version import IS_DEPRECATED_MMCV


def get_latest_ckpt(work_dir: str) -> dict:
    """This function retrieves the latest checkpoint from the given directory.

    Args:
        work_dir (str): The directory to search for checkpoints.

    Returns:
        dict: A dictionary containing the path to the latest checkpoint.
    """
    files = glob.glob(osp.join(work_dir, '*.pth'))
    if not files:
        return dict()
    return dict(path=max(files, key=os.path.getctime))


if not IS_DEPRECATED_MMCV:
    from mmengine.dist import master_only
    from mmengine.hooks import LoggerHook
    from mmengine.hooks.logger_hook import SUFFIX_TYPE
    from mmengine.registry import HOOKS
    from mmengine.runner import Runner

    @HOOKS.register_module()
    class RayTuneReporterHook(LoggerHook):
        """MMCV Logger hook for Ray Tune."""

        def __init__(self,
                     interval: int = 10,
                     ignore_last: bool = True,
                     interval_exp_name: int = 1000,
                     out_dir: Optional[Union[str, Path]] = None,
                     out_suffix: SUFFIX_TYPE = ('.json', '.log', '.py',
                                                'yaml'),
                     keep_local: bool = True,
                     file_client_args: Optional[dict] = None,
                     log_metric_by_epoch: bool = True,
                     backend_args: Optional[dict] = None,
                     filtering_key: str = 'val',
                     with_ckpt: bool = True):

            super().__init__(interval, ignore_last, interval_exp_name, out_dir,
                             out_suffix, keep_local, file_client_args,
                             log_metric_by_epoch, backend_args)
            self.filtering_key = filtering_key
            self.with_ckpt = with_ckpt

        @master_only
        def after_train_iter(self, runner: Runner, **kwargs) -> None:
            """Log after train itr.

            Args:
                runner (:obj:`mmengine.runner.Runner`): The runner to log.
            """
            batch_idx = kwargs['batch_idx']
            if self.every_n_train_iters(
                    runner, self.interval_exp_name) or (self.end_of_epoch(
                        runner.train_dataloader, batch_idx)):
                exp_info = f'Exp name: {runner.experiment_name}'
                runner.logger.info(exp_info)
            if self.every_n_inner_iters(batch_idx, self.interval):
                tag, _ = runner.log_processor.get_log_after_iter(
                    runner, batch_idx, 'train')
            elif (self.end_of_epoch(runner.train_dataloader, batch_idx)
                  and not self.ignore_last):
                # `runner.max_iters` may not be divisible by `self.interval`.
                # if `self.ignore_last==True`, the log of remaining iterations
                # will be recorded (Epoch [4][1000/1007], the logs of 998-1007
                # iterations will be recorded).
                tag, _ = runner.log_processor.get_log_after_iter(
                    runner, batch_idx, 'train')
            else:
                return
            if not any(
                    filter(lambda elem: self.filtering_key in elem,
                           tag.keys())):
                return
            ckpt = get_latest_ckpt(runner.work_dir)
            if self.with_ckpt and ckpt:
                session.report(tag, checkpoint=Checkpoint.from_dict(ckpt))
            else:
                session.report(tag)

        @master_only
        def after_val_epoch(
                self,
                runner,
                metrics: Optional[Dict[str, float]] = None) -> None:
            """All subclasses should override this method, if they need any
            operations after each validation epoch.

            Args:
                runner (Runner): The runner of the validation process.
                metrics (Dict[str, float], optional): Evaluation results of all
                    metrics on validation dataset. The keys are the names of
                    the metrics, and the values are corresponding results.
            """
            tag, _ = runner.log_processor.get_log_after_epoch(
                runner, len(runner.val_dataloader), 'val')
            tag = {k: v for k, v in tag.items() if 'time' not in k}
            if not any(
                    filter(lambda elem: self.filtering_key in elem,
                           tag.keys())):
                return
            ckpt = get_latest_ckpt(runner.work_dir)
            if self.with_ckpt and ckpt:
                session.report(tag, checkpoint=Checkpoint.from_dict(ckpt))
            else:
                session.report(tag)

else:
    from mmcv.runner import HOOKS, BaseRunner
    from mmcv.runner.dist_utils import master_only
    from mmcv.runner.hooks.logger import LoggerHook

    @HOOKS.register_module()
    class RayTuneReporterHook(LoggerHook):
        """MMCV Logger hook for Ray Tune."""

        def __init__(
            self,
            interval: int = 1,
            ignore_last: bool = True,
            reset_flag: bool = False,
            by_epoch: bool = False,
            filtering_key: str = 'val',
            with_ckpt: bool = True,
        ) -> None:
            """Initialize the hook.

            Args:
                interval (int): The interval to log.
                ignore_last (bool): Whether to ignore the last iteration.
                reset_flag (bool): Whether to reset the iteration.
                by_epoch (bool): Whether to log by epoch.
                filtering_key (str): The key to filter.
            """
            super(RayTuneReporterHook, self).__init__(interval, ignore_last,
                                                      reset_flag, by_epoch)
            self.filtering_key = filtering_key
            self.with_ckpt = with_ckpt

        def after_train_iter(self, runner: BaseRunner) -> None:
            """Log after train itr.

            Args:
                runner (:obj:`BaseRunner`): The runner to log.
            """
            if self.by_epoch and self.every_n_inner_iters(
                    runner, self.interval):
                runner.log_buffer.average(self.interval)
            elif not self.by_epoch and self.every_n_iters(
                    runner, self.interval):
                runner.log_buffer.average(self.interval)
            elif self.end_of_epoch(runner) and not self.ignore_last:
                # not precise but more stable
                runner.log_buffer.average(self.interval)

            self.log(runner)
            if self.reset_flag:
                runner.log_buffer.clear_output()

        def after_train_epoch(self, runner: BaseRunner) -> None:
            """Log after train epoch.

            Args:
                runner (:obj:`BaseRunner`): The runner to log.
            """
            self.log(runner)
            if self.reset_flag:
                runner.log_buffer.clear_output()

        def after_val_epoch(self, runner: BaseRunner) -> None:
            """Log after val epoch.

            Args:
                runner (:obj:`BaseRunner`): The runner to log.
            """
            runner.log_buffer.average()
            self.log(runner)
            if self.reset_flag:
                runner.log_buffer.clear_output()

        @master_only
        def log(self, runner: BaseRunner) -> None:
            """Log the information.

            Args:
                runner (:obj:`BaseRunner`): The runner to log.
            """

            tags = self.get_loggable_tags(runner)
            if not any(
                    filter(lambda elem: self.filtering_key in elem,
                           tags.keys())):
                return
            ckpt = get_latest_ckpt(runner.work_dir)
            if self.with_ckpt and ckpt:
                session.report(tags, checkpoint=Checkpoint.from_dict(ckpt))
            else:
                session.report(tags)



# Copyright (c) SI-Analytics. All rights reserved.
from .builder import STOPPERS, build_stopper
from .dict_stop import DictionaryStopper
from .early_drop import EarlyDroppingStopper

__all__ = [
    'STOPPERS', 'build_stopper', 'DictionaryStopper', 'EarlyDroppingStopper'
]



# Copyright (c) SI-Analytics. All rights reserved.
from mmengine.config import Config


def dump_cfg(cfg: Config, save_path: str) -> bool:
    """Dump the config. It has been split into a separate function to support
    the lower version of mmcv.

    Args:
        cfg (Config): The config to be dumped.
        save_path (str): The path to save the config.

    Returns:
        bool: Whether the config is dumped successfully.
    """
    try:
        with open(save_path, 'w', encoding='utf-8') as f:
            f.write(cfg.pretty_text)
        return True
    except Exception as err:
        from siatune.utils import get_root_logger
        logger = get_root_logger()
        logger.error(f'Failed to dump config to {save_path}: {err}')
        return False



# Copyright (c) SI-Analytics. All rights reserved.
from typing import Dict

from mmengine.registry import Registry

from .base import BaseRewriter

REWRITERS = Registry('rewriters')


def build_rewriter(cfg: Dict) -> BaseRewriter:
    """Build a rewriter.

    Args:
        cfg (Dict): The config of the rewriter.

    Returns:
        BaseRewriter: The rewriter.
    """
    return REWRITERS.build(cfg)



# Copyright (c) SI-Analytics. All rights reserved.
import re
from typing import Dict, Tuple

from .base import BaseRewriter
from .builder import REWRITERS

WRAPPING_REGEXP = r'^\$\((.*)\)$'


def unwrap_regexp(value: str,
                  regexp: str = WRAPPING_REGEXP) -> Tuple[str, bool]:
    """Unwrap the value if it is wrapped by the regexp.

    Args:
        value (str): The value to unwrap.
        regexp (str): The regexp to match.
        Defaults to WRAPPING_REGEXP.

    Returns:
        str: The unwrapped value.
        bool: Whether the value is wrapped.
    """
    if not isinstance(value, str):
        return value, False
    searched = re.search(regexp, value)
    if searched:
        value = searched.group(1)
    return value, bool(searched)


@REWRITERS.register_module()
class BatchConfigPatcher(BaseRewriter):
    """Patch the config in the context. It was introduced with the intention of
    changing two or more elements at once. To take object detection as an
    example, suppose you want to find the optimal number of fpn output
    channels.

    'base cfg'
    model = dict(
        ...,
        neck=dict(
            type='FPN',
            ...,
            out_channels=x,
        )
        roi_head=dict(
            ...,
            bbox_head=dict(
                in_channels=b,
            )
        )
    )

    If the tuning algorithm provided by ray recommends ``y``,
    we need to change the neck and head channels
    from ``x`` to ``y`` at the same time.
    It is recommended to use this class in that case.
    The usage is as follows:

    'search space'
    {
        '$(model.neck.out_channels & model.roi_head.bbox_head.in_channels)':
        dict(type='Randint', lower=a, upper=b)
    }

    'context rewriter'
    dict(type='BatchConfigPatcher', key=...),

    1. Connect the elements you want
    to change collectively with ``&`` in the search configuration.
    2. Specifies the use of BatchConfigPatcher in the context writer.
    """

    def __init__(self, key: str) -> None:
        """Initialize the rewriter.

        Args:
            key (str): The key of the config in the context.
        """
        self.key = key

    def __call__(self, context: Dict) -> Dict:
        """Batch adjustment of elements connected by &.

        Args:
            context (Dict): The context to be rewritten.

        Returns:
            Dict: The context after rewriting.
        Examples:
            inputs = {'$(a & b)': 1}
            >>> BatchConfigPatcher()(inputs)
            {'a': 1, 'b': 1}
        """
        cfg = context[self.key]

        for key, value in cfg.copy().items():
            inner_key, is_wrapped = unwrap_regexp(key)
            if not is_wrapped:
                continue
            cfg.pop(key)
            if ' & ' in inner_key:
                for sub_key in inner_key.split(' & '):
                    cfg[sub_key] = value

        return context


@REWRITERS.register_module()
class SequeunceConfigPatcher(BaseRewriter):
    """Patch the config in the context. An extended form of BatchConfigPatcher.

     The change is that the elements to be changed
     in batches do not have to have the same value.
     To take semantic segmentation as an example,
     suppose you are looking for the optimal backbone
     feature level needed to make the final prediction.
     The size of the tensor from the backbone is equal to
     [torch.Size(b_0, c_0, h_0, w_0), torch.Size(b_1, c_1, h_1, w_1), ...].

     'base cfg'
     model = dict(
         type='EncoderDecoder',
         backbone=dict(
             ...,
         )
         decode_head=dict(
             type='ASPPHead',
             in_channels=c_x,
             in_index=x,
             ...,
         ),
         ...,
     )

     If you change the in_index of the decode head,
     you will also have to change the in_inchannels to construct a valid model.
     It is recommended to use this class in that case.
     If the ray tuning algorithm recommends ``y`` as the channel index,
     the number of channels also needs to be changed from ``c_x`` to ``c_y``.
     The usage is as follows:

    'search space'
     {
         '$(model.decode_head.in_channels - model.decode_head.in_index)' =
         dict(type='Choice', categories=[[c_a, a], [c_b, b], ...])
     }
     'context rewriter'
     dict(type='SequeunceConfigPatcher', key=...),

     1. Connect the elements you want
     to change collectively with ``-`` in the search configuration.
     2. Specifies the use of SequeunceConfigPatcher in the context writer.
    """

    def __init__(self, key: str) -> None:
        """Initialize the rewriter.

        Args:
            key (str): The key of the config path in the context.
        """

        self.key = key

    def __call__(self, context: Dict) -> Dict:
        """Adjust the elements connected by - in order.

        Args:
            context (Dict): The context to be rewritten.

        Returns:
            Dict: The context after rewriting.
        Examples:
            inputs = {'$(a - b)': [1, 2]}
            >>> SequeunceConfigPatcher()(inputs)
            {'a': 1, 'b': 2}
        """
        cfg = context[self.key]

        for key, value in cfg.copy().items():
            inner_key, is_wrapped = unwrap_regexp(key)
            if not is_wrapped:
                continue
            cfg.pop(key)
            if ' - ' in inner_key:
                for idx, sub_key in enumerate(inner_key.split(' - ')):
                    cfg[sub_key] = value[idx]

        return context



# Copyright (c) SI-Analytics. All rights reserved.
__version__ = '0.4.0rc1'
from typing import Tuple

try:
    import mmcv
except ImportError:
    mmcv = None

IS_DEPRECATED_MMCV = False


def parse_version_info(version_str: str) -> Tuple:
    """Parse version string.

    Args:
        version_str (str): The version string.
    Returns:
        Tuple: The version tuple.
    """

    version_info = []
    for x in version_str.split('.'):
        if x.isdigit():
            version_info.append(int(x))
        elif x.find('rc') != -1:
            patch_version = x.split('rc')
            version_info.append(int(patch_version[0]))
            version_info.append(f'rc{patch_version[1]}')
    return tuple(version_info)


if mmcv is not None and (parse_version_info(mmcv.__version__) <
                         parse_version_info('2.0.0rc0')):
    IS_DEPRECATED_MMCV = True

version_info = parse_version_info(__version__)



# Copyright (c) SI-Analytics. All rights reserved.
import argparse
from typing import Dict

from mmengine.config import Config

from .base import BaseRewriter
from .builder import REWRITERS


@REWRITERS.register_module()
class InstantiateCfg(BaseRewriter):
    """Instantiate the configs in the argparse namespace."""
    arg_name: str = 'config'
    raw_arg_idx: int = 0

    def __init__(
        self,
        key: str,
    ) -> None:
        """Initialize the rewriter.

        Args:
            key (str): The key where the instantiated cfg is stored.
        """
        self.key = key

    def __call__(self, context: Dict) -> Dict:
        """Receive the config path from argparse namespace in the context and
        build the mmcv config file.

        Args:
            context (Dict): The context to be rewritten.

        Returns:
            Dict: The context after rewriting.
        """
        is_parsed = isinstance(context['args'], argparse.Namespace)
        file_name: str
        if is_parsed:
            file_name = getattr(context['args'], self.arg_name)
        else:
            file_name = context['args'][self.raw_arg_idx]
        context[self.key] = Config.fromfile(file_name)
        return context



# Copyright (c) SI-Analytics. All rights reserved.
import argparse
import copy
import os
import time
import warnings
from os import path as osp
from typing import Sequence

from mmengine.config import Config, ConfigDict, DictAction
from mmengine.utils import digit_version
from mmengine.utils.dl_utils import TORCH_VERSION

from siatune.version import IS_DEPRECATED_MMCV
from .builder import TASKS
from .mm import MMBaseTask

if IS_DEPRECATED_MMCV:

    @TASKS.register_module()
    class MMClassification(MMBaseTask):
        """MMClassification wrapper class for `ray.tune`.

        It is modified from https://github.com/open-mmlab/mmclassification/blob/v0.23.2/tools/train.py

        Args:
            args (argparse.Namespace): The arguments for `tools/train.py`
                script file. It is parsed by :method:`parse_args`.
            num_workers (int): The number of workers to launch.
            num_cpus_per_worker (int): The number of CPUs per worker.
                Default to 1.
            num_gpus_per_worker (int): The number of GPUs per worker.
                Since it must be equal `num_workers` attribute, it is
                not used in MMClassification.
            rewriters (list[dict] | dict, optional): Context redefinition
                pipeline. Default to None.
        """

        VERSION = 'v0.23.2'

        def parse_args(self, task_args: Sequence[str]):
            parser = argparse.ArgumentParser(description='Train a model')
            parser.add_argument('config', help='train config file path')
            parser.add_argument(
                '--work-dir', help='the dir to save logs and models')
            parser.add_argument(
                '--resume-from', help='the checkpoint file to resume from')
            parser.add_argument(
                '--no-validate',
                action='store_true',
                help='whether not to evaluate the checkpoint during training')
            group_gpus = parser.add_mutually_exclusive_group()
            group_gpus.add_argument(
                '--device', help='device used for training. (Deprecated)')
            group_gpus.add_argument(
                '--gpus',
                type=int,
                help='(Deprecated, please use --gpu-id) number of gpus to use '
                '(only applicable to non-distributed training)')
            group_gpus.add_argument(
                '--gpu-ids',
                type=int,
                nargs='+',
                help='(Deprecated, please use --gpu-id) ids of gpus to use '
                '(only applicable to non-distributed training)')
            group_gpus.add_argument(
                '--gpu-id',
                type=int,
                default=0,
                help='id of gpu to use '
                '(only applicable to non-distributed training)')
            parser.add_argument(
                '--ipu-replicas',
                type=int,
                default=None,
                help='num of ipu replicas to use')
            parser.add_argument(
                '--seed', type=int, default=None, help='random seed')
            parser.add_argument(
                '--diff-seed',
                action='store_true',
                help='Whether or not set different seeds for different ranks')
            parser.add_argument(
                '--deterministic',
                action='store_true',
                help='whether to set deterministic options for CUDNN backend.')
            parser.add_argument(
                '--cfg-options',
                nargs='+',
                action=DictAction,
                help=
                'override some settings in the used config, the key-value pair '
                'in xxx=yyy format will be merged into config file. If the value to '
                'be overwritten is a list, it should be like key="[a,b]" or key=a,b '
                'It also allows nested list/tuple values, e.g. key="[(a,b),(c,d)]" '
                'Note that the quotation marks are necessary and that no white space '
                'is allowed.')
            parser.add_argument(
                '--launcher',
                choices=['none', 'pytorch', 'slurm', 'mpi'],
                default='none',
                help='job launcher')
            parser.add_argument('--local_rank', type=int, default=0)
            args = parser.parse_args(task_args)
            if 'LOCAL_RANK' not in os.environ:
                os.environ['LOCAL_RANK'] = str(args.local_rank)

            return args

        def run(self, args: argparse.Namespace):
            """Run the task.

            Args:
                args (argparse.Namespace):
                    The args that received from context manager.
            """
            import mmcv
            import torch
            import torch.distributed as dist
            from mmcls import __version__
            from mmcls.apis import (init_random_seed, set_random_seed,
                                    train_model)
            from mmcls.datasets import build_dataset
            from mmcls.models import build_classifier
            from mmcls.utils import (auto_select_device, collect_env,
                                     get_root_logger, setup_multi_processes)
            from mmcv import Config
            from mmcv.runner import get_dist_info, init_dist

            cfg = Config.fromfile(args.config)
            if args.cfg_options is not None:
                cfg.merge_from_dict(args.cfg_options)

            # set multi-process settings
            setup_multi_processes(cfg)

            # set cudnn_benchmark
            if cfg.get('cudnn_benchmark', False):
                torch.backends.cudnn.benchmark = True

            # work_dir is determined in this priority: CLI > segment in file > filename
            if args.work_dir is not None:
                # update configs according to CLI args if args.work_dir is not None
                cfg.work_dir = args.work_dir
            elif cfg.get('work_dir', None) is None:
                # use config filename as default work_dir if cfg.work_dir is None
                cfg.work_dir = osp.join(
                    './work_dirs',
                    osp.splitext(osp.basename(args.config))[0])
            if args.resume_from is not None:
                cfg.resume_from = args.resume_from
            if args.gpus is not None:
                cfg.gpu_ids = range(1)
                warnings.warn('`--gpus` is deprecated because we only support '
                              'single GPU mode in non-distributed training. '
                              'Use `gpus=1` now.')
            if args.gpu_ids is not None:
                cfg.gpu_ids = args.gpu_ids[0:1]
                warnings.warn(
                    '`--gpu-ids` is deprecated, please use `--gpu-id`. '
                    'Because we only support single GPU mode in '
                    'non-distributed training. Use the first GPU '
                    'in `gpu_ids` now.')
            if args.gpus is None and args.gpu_ids is None:
                cfg.gpu_ids = [args.gpu_id]

            if args.ipu_replicas is not None:
                cfg.ipu_replicas = args.ipu_replicas
                args.device = 'ipu'

            # init distributed env first, since logger depends on the dist info.
            if args.launcher == 'none':
                distributed = False
            else:
                distributed = True
                init_dist(args.launcher, **cfg.dist_params)
                _, world_size = get_dist_info()
                cfg.gpu_ids = range(world_size)

            # create work_dir
            mmcv.mkdir_or_exist(osp.abspath(cfg.work_dir))
            # dump config
            cfg.dump(osp.join(cfg.work_dir, osp.basename(args.config)))
            # init the logger before other steps
            timestamp = time.strftime('%Y%m%d_%H%M%S', time.localtime())
            log_file = osp.join(cfg.work_dir, f'{timestamp}.log')
            logger = get_root_logger(
                log_file=log_file, log_level=cfg.log_level)

            # init the meta dict to record some important information such as
            # environment info and seed, which will be logged
            meta = dict()
            # log env info
            env_info_dict = collect_env()
            env_info = '\n'.join([(f'{k}: {v}')
                                  for k, v in env_info_dict.items()])
            dash_line = '-' * 60 + '\n'
            logger.info('Environment info:\n' + dash_line + env_info + '\n' +
                        dash_line)
            meta['env_info'] = env_info

            # log some basic info
            logger.info(f'Distributed training: {distributed}')
            logger.info(f'Config:\n{cfg.pretty_text}')

            # set random seeds
            cfg.device = args.device or auto_select_device()
            seed = init_random_seed(args.seed, device=cfg.device)
            seed = seed + dist.get_rank() if args.diff_seed else seed
            logger.info(f'Set random seed to {seed}, '
                        f'deterministic: {args.deterministic}')
            set_random_seed(seed, deterministic=args.deterministic)
            cfg.seed = seed
            meta['seed'] = seed

            model = build_classifier(cfg.model)
            model.init_weights()

            datasets = [build_dataset(cfg.data.train)]
            if len(cfg.workflow) == 2:
                val_dataset = copy.deepcopy(cfg.data.val)
                val_dataset.pipeline = cfg.data.train.pipeline
                datasets.append(build_dataset(val_dataset))

            # save mmcls version, config file content and class names in
            # runner as meta data
            meta.update(
                dict(
                    mmcls_version=__version__,
                    config=cfg.pretty_text,
                    CLASSES=datasets[0].CLASSES))

            # add an attribute for visualization convenience
            train_model(
                model,
                datasets,
                cfg,
                distributed=distributed,
                validate=(not args.no_validate),
                timestamp=timestamp,
                device=cfg.device,
                meta=meta)
else:

    @TASKS.register_module()
    class MMClassification(MMBaseTask):
        """MMClassification wrapper class for `ray.tune`.

        It is modified from https://github.com/open-mmlab/mmclassification/blob/v1.0.0rc4/tools/train.py

        Args:
            args (argparse.Namespace): The arguments for `tools/train.py`
                script file. It is parsed by :method:`parse_args`.
            num_workers (int): The number of workers to launch.
            num_cpus_per_worker (int): The number of CPUs per worker.
                Default to 1.
            num_gpus_per_worker (int): The number of GPUs per worker.
                Since it must be equal `num_workers` attribute, it is
                not used in MMClassification.
        """

        VERSION = 'v1.0.0rc4'

        def parse_args(self, task_args: Sequence[str]):
            parser = argparse.ArgumentParser(description='Train a classifier')
            parser.add_argument('config', help='train config file path')
            parser.add_argument(
                '--work-dir', help='the dir to save logs and models')
            parser.add_argument(
                '--resume',
                nargs='?',
                type=str,
                const='auto',
                help='If specify checkpoint path, resume from it, while if not '
                'specify, try to auto resume from the latest checkpoint '
                'in the work directory.')
            parser.add_argument(
                '--amp',
                action='store_true',
                help='enable automatic-mixed-precision training')
            parser.add_argument(
                '--no-validate',
                action='store_true',
                help='whether not to evaluate the checkpoint during training')
            parser.add_argument(
                '--auto-scale-lr',
                action='store_true',
                help='whether to auto scale the learning rate according to the '
                'actual batch size and the original batch size.')
            parser.add_argument(
                '--no-pin-memory',
                action='store_true',
                help='whether to disable the pin_memory option in dataloaders.'
            )
            parser.add_argument(
                '--no-persistent-workers',
                action='store_true',
                help=
                'whether to disable the persistent_workers option in dataloaders.'
            )
            parser.add_argument(
                '--cfg-options',
                nargs='+',
                action=DictAction,
                help=
                'override some settings in the used config, the key-value pair '
                'in xxx=yyy format will be merged into config file. If the value to '
                'be overwritten is a list, it should be like key="[a,b]" or key=a,b '
                'It also allows nested list/tuple values, e.g. key="[(a,b),(c,d)]" '
                'Note that the quotation marks are necessary and that no white space '
                'is allowed.')
            parser.add_argument(
                '--launcher',
                choices=['none', 'pytorch', 'slurm', 'mpi'],
                default='none',
                help='job launcher')
            parser.add_argument('--local_rank', type=int, default=0)
            args = parser.parse_args(task_args)
            if 'LOCAL_RANK' not in os.environ:
                os.environ['LOCAL_RANK'] = str(args.local_rank)

            return args

        def run(self, args: argparse.Namespace):
            """Run the task.

            Args:
                args (argparse.Namespace):
                    The args that received from context manager.
            """
            from copy import deepcopy

            from mmcls.utils import register_all_modules
            from mmengine.runner import Runner

            def merge_args(cfg, args):
                """Merge CLI arguments to config."""
                if args.no_validate:
                    cfg.val_cfg = None
                    cfg.val_dataloader = None
                    cfg.val_evaluator = None

                cfg.launcher = args.launcher

                # work_dir is determined in this priority: CLI > segment in file > filename
                if args.work_dir is not None:
                    # update configs according to CLI args if args.work_dir is not None
                    cfg.work_dir = args.work_dir
                elif cfg.get('work_dir', None) is None:
                    # use config filename as default work_dir if cfg.work_dir is None
                    cfg.work_dir = osp.join(
                        './work_dirs',
                        osp.splitext(osp.basename(args.config))[0])

                # enable automatic-mixed-precision training
                if args.amp is True:
                    optim_wrapper = cfg.optim_wrapper.get(
                        'type', 'OptimWrapper')
                    assert optim_wrapper in ['OptimWrapper', 'AmpOptimWrapper'], \
                        '`--amp` is not supported custom optimizer wrapper type ' \
                        f'`{optim_wrapper}.'
                    cfg.optim_wrapper.type = 'AmpOptimWrapper'
                    cfg.optim_wrapper.setdefault('loss_scale', 'dynamic')

                # resume training
                if args.resume == 'auto':
                    cfg.resume = True
                    cfg.load_from = None
                elif args.resume is not None:
                    cfg.resume = True
                    cfg.load_from = args.resume

                # enable auto scale learning rate
                if args.auto_scale_lr:
                    cfg.auto_scale_lr.enable = True

                # set dataloader args
                default_dataloader_cfg = ConfigDict(
                    pin_memory=True,
                    persistent_workers=True,
                    collate_fn=dict(type='default_collate'),
                )
                if digit_version(TORCH_VERSION) < digit_version('1.8.0'):
                    default_dataloader_cfg.persistent_workers = False

                def set_default_dataloader_cfg(cfg, field):
                    if cfg.get(field, None) is None:
                        return
                    dataloader_cfg = deepcopy(default_dataloader_cfg)
                    dataloader_cfg.update(cfg[field])
                    cfg[field] = dataloader_cfg
                    if args.no_pin_memory:
                        cfg[field]['pin_memory'] = False
                    if args.no_persistent_workers:
                        cfg[field]['persistent_workers'] = False

                set_default_dataloader_cfg(cfg, 'train_dataloader')
                set_default_dataloader_cfg(cfg, 'val_dataloader')
                set_default_dataloader_cfg(cfg, 'test_dataloader')

                if args.cfg_options is not None:
                    cfg.merge_from_dict(args.cfg_options)

                return cfg

            # register all modules in mmcls into the registries
            # do not init the default scope here because it will be init in the runner
            register_all_modules(init_default_scope=False)

            # load config
            cfg = Config.fromfile(args.config)

            # merge cli arguments to config
            cfg = merge_args(cfg, args)

            # build the runner from config
            runner = Runner.from_cfg(cfg)

            # start training
            runner.train()



# Copyright (c) SI-Analytics. All rights reserved.

from mmengine.config import Config
from mmengine.registry import Registry
from ray import tune
from ray.tune.search import Searcher

SEARCHERS = Registry('searcher')

# Dynamically import search_alg
# Refer to https://github.com/ray-project/ray/blob/master/python/ray/tune/search/__init__.py  # noqa
for cls in set(func() for func in tune.search.SEARCH_ALG_IMPORT.values()):
    SEARCHERS.register_module(module=cls)


def build_searcher(cfg: Config) -> Searcher:
    """Build the searcher from configs.

    Args:
        cfg (Config): The configs.

    Returns:
        tune.suggest.Searcher: The searcher.
    """

    return SEARCHERS.build(cfg)



# Copyright (c) SI-Analytics. All rights reserved.
import copy
from typing import Any, Optional


class _Freezer(object):
    """Freeze any class, such that instantiated objects become immutable.

    https://medium.datadriveninvestor.com/immutability-in-python-d57a3b23f336
    """

    _lock = False

    def __init__(self):
        """Initialize the freezer."""
        self._lock = True

    def __delattr__(self, *args, **kwargs) -> None:
        """Delete an attribute.

        Raises:
            AttributeError: If the object is locked.
        """

        if self._lock:
            raise AttributeError
        object.__delattr__(self, *args, **kwargs)

    def __setattr__(self, *args, **kwargs) -> None:
        """Set an attribute.

        Raises:
            AttributeError: If the object is locked.
        """

        if self._lock:
            raise AttributeError
        object.__setattr__(self, *args, **kwargs)


class ImmutableContainer(_Freezer):
    """Ensure object immutability."""

    MAX_REPR_LEN: int = 16

    def __init__(self, data: Any, alias: Optional[str] = None) -> None:
        """Initialize the container.

        Args:
            data (Any): The data to be stored.
            alias (Optional[str]): The alias of the data.
        """
        self.__data = data
        self._alias = alias
        super().__init__()

    def __repr__(self) -> str:
        """Return the string representation of the container."""
        if self._alias is not None:
            return self._alias
        elif len(self.__data.__repr__()) > ImmutableContainer.MAX_REPR_LEN:
            return f'{self.__class__.__name__}( * )'
        return f'{self.__class__.__name__}( {repr(self.data)} )'

    @property
    def data(self):
        """Return the data."""
        return copy.deepcopy(self.__data)

    @data.setter
    def data(self):
        raise AttributeError(
            'Setting data inside an immutable container is not allowed.')

    @property
    def alias(self):
        """Return the alias."""
        return self._alias

    def __eq__(self, other: Any) -> bool:
        """Return True if the containers are equal.

        Args:
            other (Any): The other object to compare.
        Returns:
            bool: True if the other object is equal.
        """
        if isinstance(other, ImmutableContainer):
            return self.data == other.data
        return self.data == other

    def __hash__(self) -> int:
        """Return hash value of `str(self.data)`. It needs to use FLAML.

        Returns:
            int: hash value.
        """
        return hash(str(self.data))

    @classmethod
    def decouple(cls, inputs: Any) -> Any:
        """Decouple the inputs.

        Args:
            inputs (Any): The inputs to be decoupled.

        Returns:
            Any: The decoupled inputs.
        """
        if isinstance(inputs, ImmutableContainer):
            return inputs.data
        elif isinstance(inputs, dict):
            outputs = inputs.copy()
            for k, v in inputs.items():
                outputs[k] = ImmutableContainer.decouple(v)
            return outputs
        elif isinstance(inputs, list):
            outputs = inputs.copy()
            for idx, elm in enumerate(inputs):
                outputs[idx] = ImmutableContainer.decouple(elm)
            return outputs
        return inputs



# Copyright (c) SI-Analytics. All rights reserved.
from abc import ABCMeta
from typing import Callable

import ray.tune as tune
from ray.tune.search.sample import Domain

from .builder import SPACES


class BaseSpace(metaclass=ABCMeta):
    """Base Space class."""
    sample: Callable = None

    def __init__(self, **kwargs) -> None:
        self.kwargs = kwargs

    @property
    def space(self) -> Domain:
        """Return the space."""
        return self.sample.__func__(**self.kwargs)


@SPACES.register_module()
class Uniform(BaseSpace):
    sample: Callable = tune.uniform


@SPACES.register_module()
class Quniform(BaseSpace):
    sample: Callable = tune.quniform


@SPACES.register_module()
class Loguniform(BaseSpace):
    sample: Callable = tune.loguniform


@SPACES.register_module()
class Qloguniform(BaseSpace):
    sample: Callable = tune.qloguniform


@SPACES.register_module()
class Randn(BaseSpace):
    sample: Callable = tune.randn


@SPACES.register_module()
class Qrandn(BaseSpace):
    sample: Callable = tune.qrandn


@SPACES.register_module()
class Randint(BaseSpace):
    sample: Callable = tune.randint


@SPACES.register_module()
class Qrandint(BaseSpace):
    sample: Callable = tune.qrandint


@SPACES.register_module()
class Lograndint(BaseSpace):
    sample: Callable = tune.lograndint


@SPACES.register_module()
class Qlograndint(BaseSpace):
    sample: Callable = tune.qlograndint



# Copyright (c) SI-Analytics. All rights reserved.
from .base import BaseRewriter
from .builder import REWRITERS, build_rewriter
from .dump import Dump
from .instantiate import InstantiateCfg
from .merge import MergeConfig
from .patch import BatchConfigPatcher, SequeunceConfigPatcher
from .path import AttachTrialInfoToPath
from .register import CustomHookRegister
from .resume import ResumeFromCkpt

__all__ = [
    'BaseRewriter', 'REWRITERS', 'build_rewriter', 'Dump', 'MergeConfig',
    'AppendTrialIDtoPath', 'BatchConfigPatcher', 'SequeunceConfigPatcher',
    'CustomHookRegister', 'InstantiateCfg', 'ResumeFromCkpt',
    'AttachTrialInfoToPath'
]



# Copyright (c) SI-Analytics. All rights reserved.
import argparse
from os import path as osp

from ray.air import session

from siatune.utils import reference_raw_args
from .base import BaseRewriter
from .builder import REWRITERS


@REWRITERS.register_module()
class AttachTrialInfoToPath(BaseRewriter):
    """Add the identifier of the tials to the workspace path to prevent the
    artifacts of each trial from being stored in the same path."""
    arg_name: str = 'work_dir'
    raw_arg_name: str = '--work-dir'

    def __call__(self, context: dict) -> dict:
        """Give the workspace a different ID for each trial.

        Args:
            context (dict): The context to be rewritten.

        Returns:
            dict: The context after rewriting.
        """
        is_parsed = isinstance(context['args'], argparse.Namespace)
        if is_parsed:
            work_dir = getattr(context['args'], self.arg_name, '')
            if work_dir:
                work_dir = osp.join(work_dir, session.get_trial_id())
            else:
                work_dir = session.get_trial_dir()
            setattr(context['args'], self.arg_name, work_dir)
        else:
            work_dir, idx = reference_raw_args(context['args'],
                                               self.raw_arg_name)
            if idx:
                context['args'][idx.pop()] = osp.join(work_dir.pop(),
                                                      session.get_trial_id())
            else:
                context['args'].extend([
                    self.raw_arg_name,
                    osp.join('.', session.get_trial_dir())
                ])
        return context



# Copyright (c) SI-Analytics. All rights reserved.
from abc import ABCMeta
from copy import deepcopy
from typing import Callable

from ray.tune import with_resources

from siatune.core import ContextManager, DistributedTorchLauncher
from siatune.utils import ImmutableContainer
from .base import BaseTask
from .builder import TASKS


@TASKS.register_module()
class MMBaseTask(BaseTask, metaclass=ABCMeta):
    """Wrap the apis of open mm train-based projects."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        assert self.num_gpus_per_worker == 1
        self.launcher = DistributedTorchLauncher(
            self.num_cpus_per_worker,
            self.num_workers,
        )

    def create_trainable(self) -> Callable:
        """Get a trainable task.

        Returns:
            Callable: Callable object to optimize hyperparameter.
        """
        return with_resources(self.context_aware_run, self.launcher.resources)

    def context_aware_run(self, searched_cfg: dict):
        """Gather and refine the information received by users and Ray.tune to
        execute the objective task.

        Args:
            searched_cfg (Dict): The searched configuration.
        """

        context_manager = ContextManager(self.rewriters)
        context = dict(
            args=deepcopy(self.args),
            searched_cfg=deepcopy(ImmutableContainer.decouple(searched_cfg)),
        )
        return context_manager(self.dist_run)(**context)

    def dist_run(self, *args, **kwargs):
        self.launcher.launch(self.run, *args, **kwargs)



# Copyright (c) SI-Analytics. All rights reserved.
import collections
from typing import List

from .rewriters.builder import build_rewriter


# TODO: Use a context manager as decorator.
class ContextManager:
    """The context manager receives the context from the user and the ray
    tuning algorithm, and refines it into a form usable by the task
    processor."""

    def __init__(self, rewriters: List[dict] = []) -> None:
        """initialize the context manager.

        Args:
            rewriters (List[dict]):
                User-defined context rewriting pipeline.
                Defaults to [].

        Raises:
            TypeError: If the rewriters are not a list.
        """
        self.rewriters = []
        assert isinstance(rewriters, collections.abc.Sequence)
        for rewriter in rewriters:
            if isinstance(rewriter, dict):
                rewriter = build_rewriter(rewriter)
                self.rewriters.append(rewriter)
            elif callable(rewriter):
                self.rewriters.append(rewriter)
            else:
                raise TypeError('rewriter must be callable or a dict')

    def __call__(self, func: callable) -> callable:
        """rewrite the context.

        Args:
            func (callable): The function to be decorated.
        """

        def inner(**context):

            for rewriter in self.rewriters:
                context = rewriter(context)
            return func(**context)

        return inner



# Copyright (c) SI-Analytics. All rights reserved.
import argparse
import copy
import logging
import os
import time
from os import path as osp
from typing import Sequence

from mmengine.config import Config, DictAction

from siatune.version import IS_DEPRECATED_MMCV
from .builder import TASKS
from .mm import MMBaseTask

if IS_DEPRECATED_MMCV:

    @TASKS.register_module()
    class MMEditing(MMBaseTask):
        """MMEditing wrapper class for `ray.tune`.

        It is modified from https://github.com/open-mmlab/mmediting/blob/v0.15.0/tools/train.py

        Args:
            args (argparse.Namespace): The arguments for `tools/train.py`
                script file. It is parsed by :method:`parse_args`.
            num_workers (int): The number of workers to launch.
            num_cpus_per_worker (int): The number of CPUs per worker.
                Default to 1.
            num_gpus_per_worker (int): The number of GPUs per worker.
                Since it must be equal to `num_workers` attribute, it
                is not used in MMEditing.
            rewriters (list[dict] | dict, optional): Context redefinition
                pipeline. Default to None.
        """

        VERSION = 'v0.15.0'

        def parse_args(self, task_args: Sequence[str]):
            from mmcv import DictAction

            parser = argparse.ArgumentParser(description='Train an editor')
            parser.add_argument('config', help='train config file path')
            parser.add_argument(
                '--work-dir', help='the dir to save logs and models')
            parser.add_argument(
                '--resume-from', help='the checkpoint file to resume from')
            parser.add_argument(
                '--no-validate',
                action='store_true',
                help='whether not to evaluate the checkpoint during training')
            parser.add_argument(
                '--gpus',
                type=int,
                default=1,
                help='number of gpus to use '
                '(only applicable to non-distributed training)')
            parser.add_argument(
                '--seed', type=int, default=None, help='random seed')
            parser.add_argument(
                '--diff_seed',
                action='store_true',
                help='Whether or not set different seeds for different ranks')
            parser.add_argument(
                '--deterministic',
                action='store_true',
                help='whether to set deterministic options for CUDNN backend.')
            parser.add_argument(
                '--cfg-options',
                nargs='+',
                action=DictAction,
                help=
                'override some settings in the used config, the key-value pair '
                'in xxx=yyy format will be merged into config file. If the value to '
                'be overwritten is a list, it should be like key="[a,b]" or key=a,b '
                'It also allows nested list/tuple values, e.g. key="[(a,b),(c,d)]" '
                'Note that the quotation marks are necessary and that no white space '
                'is allowed.')
            parser.add_argument(
                '--launcher',
                choices=['none', 'pytorch', 'slurm', 'mpi'],
                default='none',
                help='job launcher')
            parser.add_argument('--local_rank', type=int, default=0)
            parser.add_argument(
                '--autoscale-lr',
                action='store_true',
                help='automatically scale lr with the number of gpus')
            args = parser.parse_args(task_args)
            if 'LOCAL_RANK' not in os.environ:
                os.environ['LOCAL_RANK'] = str(args.local_rank)

            return args

        def run(self, args: argparse.Namespace):
            """Run the task.

            Args:
                args (argparse.Namespace):
                    The args that received from context manager.
            """

            import mmcv
            import torch
            import torch.distributed as dist
            from mmcv import Config
            from mmcv.runner import init_dist
            from mmedit import __version__
            from mmedit.apis import (init_random_seed, set_random_seed,
                                     train_model)
            from mmedit.datasets import build_dataset
            from mmedit.models import build_model
            from mmedit.utils import (collect_env, get_root_logger,
                                      setup_multi_processes)

            cfg = Config.fromfile(args.config)

            if args.cfg_options is not None:
                cfg.merge_from_dict(args.cfg_options)

            # set multi-process settings
            setup_multi_processes(cfg)

            # set cudnn_benchmark
            if cfg.get('cudnn_benchmark', False):
                torch.backends.cudnn.benchmark = True
            # update configs according to CLI args
            if args.work_dir is not None:
                cfg.work_dir = args.work_dir
            if args.resume_from is not None:
                cfg.resume_from = args.resume_from
            cfg.gpus = args.gpus

            if args.autoscale_lr:
                # apply the linear scaling rule (https://arxiv.org/abs/1706.02677)
                cfg.optimizer['lr'] = cfg.optimizer['lr'] * cfg.gpus / 8

            # init distributed env first, since logger depends on the dist info.
            if args.launcher == 'none':
                distributed = False
            else:
                distributed = True
                init_dist(args.launcher, **cfg.dist_params)

            # create work_dir
            mmcv.mkdir_or_exist(osp.abspath(cfg.work_dir))
            # init the logger before other steps
            timestamp = time.strftime('%Y%m%d_%H%M%S', time.localtime())
            log_file = osp.join(cfg.work_dir, f'{timestamp}.log')
            logger = get_root_logger(
                log_file=log_file, log_level=cfg.log_level)

            # log env info
            env_info_dict = collect_env.collect_env()
            env_info = '\n'.join(
                [f'{k}: {v}' for k, v in env_info_dict.items()])
            dash_line = '-' * 60 + '\n'
            logger.info('Environment info:\n' + dash_line + env_info + '\n' +
                        dash_line)

            # log some basic info
            logger.info('Distributed training: {}'.format(distributed))
            logger.info('mmedit Version: {}'.format(__version__))
            logger.info('Config:\n{}'.format(cfg.text))

            # set random seeds
            seed = init_random_seed(args.seed)
            seed = seed + dist.get_rank() if args.diff_seed else seed
            logger.info('Set random seed to {}, deterministic: {}'.format(
                seed, args.deterministic))
            set_random_seed(seed, deterministic=args.deterministic)
            cfg.seed = seed

            model = build_model(
                cfg.model, train_cfg=cfg.train_cfg, test_cfg=cfg.test_cfg)

            datasets = [build_dataset(cfg.data.train)]
            if len(cfg.workflow) == 2:
                val_dataset = copy.deepcopy(cfg.data.val)
                val_dataset.pipeline = cfg.data.train.pipeline
                datasets.append(build_dataset(val_dataset))
            if cfg.checkpoint_config is not None:
                # save version, config file content and class names in
                # checkpoints as meta data
                cfg.checkpoint_config.meta = dict(
                    mmedit_version=__version__,
                    config=cfg.text,
                )

            # meta information
            meta = dict()
            if cfg.get('exp_name', None) is None:
                cfg['exp_name'] = osp.splitext(osp.basename(cfg.work_dir))[0]
            meta['exp_name'] = cfg.exp_name
            meta['mmedit Version'] = __version__
            meta['seed'] = seed
            meta['env_info'] = env_info

            # add an attribute for visualization convenience
            train_model(
                model,
                datasets,
                cfg,
                distributed=distributed,
                validate=(not args.no_validate),
                timestamp=timestamp,
                meta=meta)

else:

    @TASKS.register_module()
    class MMEditing(MMBaseTask):
        """MMEditing wrapper class for `ray.tune`.

        It is modified from https://github.com/open-mmlab/mmediting/blob/v1.0.0rc4/tools/train.py

        Args:
            args (argparse.Namespace): The arguments for `tools/train.py`
                script file. It is parsed by :method:`parse_args`.
            num_workers (int): The number of workers to launch.
            num_cpus_per_worker (int): The number of CPUs per worker.
                Default to 1.
            num_gpus_per_worker (int): The number of GPUs per worker.
                Since it must be equal to `num_workers` attribute, it
                is not used in MMEditing.
            rewriters (list[dict] | dict, optional): Context redefinition
                pipeline. Default to None.
        """

        VERSION = 'v1.0.0rc4'

        def parse_args(self, task_args: Sequence[str]):
            parser = argparse.ArgumentParser(description='Train a model')
            parser.add_argument('config', help='train config file path')
            parser.add_argument(
                '--work-dir', help='the dir to save logs and models')
            parser.add_argument(
                '--resume',
                action='store_true',
                help='Whether to resume checkpoint.')
            parser.add_argument(
                '--amp',
                action='store_true',
                default=False,
                help='enable automatic-mixed-precision training')
            parser.add_argument(
                '--auto-scale-lr',
                action='store_true',
                help='enable automatically scaling LR.')
            parser.add_argument(
                '--cfg-options',
                nargs='+',
                action=DictAction,
                help=
                'override some settings in the used config, the key-value pair '
                'in xxx=yyy format will be merged into config file. If the value to '
                'be overwritten is a list, it should be like key="[a,b]" or key=a,b '
                'It also allows nested list/tuple values, e.g. key="[(a,b),(c,d)]" '
                'Note that the quotation marks are necessary and that no white space '
                'is allowed.')
            parser.add_argument(
                '--launcher',
                choices=['none', 'pytorch', 'slurm', 'mpi'],
                default='none',
                help='job launcher')
            parser.add_argument('--local_rank', type=int, default=0)
            args = parser.parse_args(task_args)
            if 'LOCAL_RANK' not in os.environ:
                os.environ['LOCAL_RANK'] = str(args.local_rank)

            return args

        def run(self, args: argparse.Namespace):
            """Run the task.

            Args:
                args (argparse.Namespace):
                    The args that received from context manager.
            """
            from mmedit.utils import print_colored_log, register_all_modules
            from mmengine.runner import Runner

            # register all modules in mmedit into the registries
            # do not init the default scope here because it will be init in the runner
            register_all_modules(init_default_scope=False)

            # load config
            cfg = Config.fromfile(args.config)
            cfg.launcher = args.launcher
            if args.cfg_options is not None:
                cfg.merge_from_dict(args.cfg_options)

            # work_dir is determined in this priority: CLI > segment in file > filename
            if args.work_dir:  # none or empty str
                # update configs according to CLI args if args.work_dir is not None
                cfg.work_dir = args.work_dir
            elif cfg.get('work_dir', None) is None:
                # use config filename as default work_dir if cfg.work_dir is None
                cfg.work_dir = osp.join(
                    './work_dirs',
                    osp.splitext(osp.basename(args.config))[0])

            # enable automatic-mixed-precision training
            if args.amp is True:
                if ('constructor' not in cfg.optim_wrapper) or \
                        cfg.optim_wrapper['constructor'] == 'DefaultOptimWrapperConstructor': # noqa
                    optim_wrapper = cfg.optim_wrapper.type
                    if optim_wrapper == 'AmpOptimWrapper':
                        print_colored_log(
                            'AMP training is already enabled in your config.',
                            logger='current',
                            level=logging.WARNING)
                    else:
                        assert optim_wrapper == 'OptimWrapper', (
                            '`--amp` is only supported when the optimizer wrapper '
                            f'`type is OptimWrapper` but got {optim_wrapper}.')
                        cfg.optim_wrapper.type = 'AmpOptimWrapper'
                        cfg.optim_wrapper.loss_scale = 'dynamic'
                else:
                    for key, val in cfg.optim_wrapper.items():
                        if isinstance(val, dict) and 'type' in val:
                            assert val.type == 'OptimWrapper', (
                                '`--amp` is only supported when the optimizer wrapper '
                                f'`type is OptimWrapper` but got {val.type}.')
                            cfg.optim_wrapper[key].type = 'AmpOptimWrapper'
                            cfg.optim_wrapper[key].loss_scale = 'dynamic'

            if args.resume:
                cfg.resume = True

            # build the runner from config
            runner = Runner.from_cfg(cfg)

            print_colored_log(f'Working directory: {cfg.work_dir}')
            print_colored_log(f'Log directiry: {runner._log_dir}')

            # start training
            runner.train()

            print_colored_log(f'Log saved under {runner._log_dir}')
            print_colored_log(f'Checkpoint saved under {cfg.work_dir}')



# Copyright (c) SI-Analytics. All rights reserved.
import os
import shutil
from os import path as osp
from pprint import pformat
from typing import Optional

import mmengine
from ray.tune import ResultGrid

from siatune.utils import ImmutableContainer
from siatune.utils.logger import get_root_logger


def log_analysis(results: ResultGrid, log_dir: Optional[str] = None) -> None:
    """Log the analysis of the experiment.

    Args:
        results (ResultGrid): Experiment results of `Tuner.tune()`.
        log_dir (str, optional): The log dir. Defaults to None.
    """

    log_dir = osp.join(log_dir or os.getcwd(), 'best_trial')
    mmengine.mkdir_or_exist(log_dir)

    logger = get_root_logger(log_file=osp.join(log_dir, 'best_trial.log'))
    result = results.get_best_result()

    logger.info(f'Best Logdir: {result.log_dir}')
    logger.info(f'Best Result: \n'
                f'{pformat(ImmutableContainer.decouple(result))}')
    logger.info(f'Best Hyperparam: \n'
                f'{pformat(ImmutableContainer.decouple(result.config))}')

    shutil.copytree(result.log_dir, osp.join(log_dir, 'log'))
